#####################################################################################
# The MIT License (MIT)
#
# Copyright (c) 2015-2024 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#####################################################################################
import cppcheck, itertools
from cppcheckdata import simpleMatch, match


def skipTokenMatches(tokens, skip=None):
    for tok in tokens:
        if match(tok, skip):
            continue
        yield tok


def isTokensEqual(xtokens, ytokens, skip=None):
    for x, y in itertools.zip_longest(skipTokenMatches(xtokens, skip),
                                      skipTokenMatches(ytokens, skip)):
        if not x:
            return False
        if not y:
            return False
        if x.str != y.str:
            return False
    return True


def getInnerLink(token):
    if not token:
        return []
    if not token.link:
        return []
    return token.next.forward(token.link)


def getVariableDecl(var):
    if not var:
        return []
    end = var.typeEndToken
    if end:
        end = end.next
    return var.typeStartToken.forward(end)


def isFunctionCall(token):
    if not token:
        return False
    if not token.isName:
        return False
    if not token.next:
        return False
    if not token.next.str == '(':
        return False
    return True


@cppcheck.checker
def AvoidBranchingStatementAsLastInLoop(cfg, data):
    for token in cfg.tokenlist:
        if not token.str in ['for', 'while']:
            continue
        end = match(token, "for|while (*) {*}").end
        if not end:
            continue
        stmt = end.tokAt(-2)
        if not match(stmt, "%any% ; }"):
            continue
        if not match(stmt, "break|continue"):
            stmt = stmt.astTop()
        if match(stmt, "break|continue|return"):
            cppcheck.reportError(
                stmt, "style",
                "Branching statement as the last statement inside a loop is very confusing."
            )


# @cppcheck.checker
# def CollapsibleIfStatements(cfg, data):
#     for token in cfg.tokenlist:
#         if not match(token, "if (*) { if (*) {*} }"):
#             continue
#         cppcheck.reportError(token, "style", "These two if statements can be collapsed into one.")


@cppcheck.checker
def ConditionalAssert(cfg, data):
    for token in cfg.tokenlist:
        if token.str != 'if':
            continue
        if not match(token, "if (*) { assert (*) ; }"):
            continue
        cppcheck.reportError(token, "style",
                             "The if condition should be included in assert.")


@cppcheck.checker
def EmptyCatchStatement(cfg, data):
    for token in cfg.tokenlist:
        if token.str != 'catch':
            continue
        if not match(token, "catch (*) { }"):
            continue
        cppcheck.reportError(token, "style", "An empty catch statement.")


@cppcheck.checker
def EmptyDoWhileStatement(cfg, data):
    for token in cfg.tokenlist:
        if token.str != 'do':
            continue
        if not simpleMatch(token, "do { } while ("):
            continue
        cppcheck.reportError(token, "style", "Empty do-while.")


@cppcheck.checker
def EmptyElseBlock(cfg, data):
    for token in cfg.tokenlist:
        if token.str != 'else':
            continue
        if not simpleMatch(token, "else { }"):
            continue
        cppcheck.reportError(token, "style",
                             "Empty else statement can be safely removed.")


@cppcheck.checker
def EmptyForStatement(cfg, data):
    for token in cfg.tokenlist:
        if token.str != 'for':
            continue
        if not match(token, "for (*) { }"):
            continue
        cppcheck.reportError(token, "style", "Empty for statement.")


@cppcheck.checker
def EmptyIfStatement(cfg, data):
    for token in cfg.tokenlist:
        if token.str != 'if':
            continue
        if not match(token, "if (*) { }"):
            continue
        cppcheck.reportError(token, "style", "Empty if statement.")


@cppcheck.checker
def EmptySwitchStatement(cfg, data):
    for token in cfg.tokenlist:
        if token.str != 'switch':
            continue
        if not match(token, "switch (*) { }"):
            continue
        cppcheck.reportError(token, "style", "Empty switch statement.")


@cppcheck.checker
def EmptyWhileStatement(cfg, data):
    for token in cfg.tokenlist:
        if token.str != 'while':
            continue
        if not match(token, "while (*) { }"):
            continue
        cppcheck.reportError(token, "style", "Empty while statement.")


@cppcheck.checker
def ForLoopShouldBeWhileLoop(cfg, data):
    for token in cfg.tokenlist:
        if token.str != 'for':
            continue
        if not match(token, "for ( ; !!;"):
            continue
        # Skip empty for loops
        if match(token, "for (*) { }"):
            continue
        end = token.next.link
        if not match(end.tokAt(-1), "; )"):
            continue
        cppcheck.reportError(token, "style",
                             "For loop should be written as a while loop.")


@cppcheck.checker
def GotoStatement(cfg, data):
    for token in cfg.tokenlist:
        if token.str != 'goto':
            continue
        cppcheck.reportError(token, "style", "Goto considered harmful.")


# @cppcheck.checker
# def InvertedLogic(cfg, data):
#     for token in cfg.tokenlist:
#         cond = None
#         if match(token, "if (*) {*} else { !!if"):
#             cond = token.next.astOperand2
#         elif match(token, "?"):
#             cond = token.astOperand1
#         if not match(cond, "!|!="):
#             continue
#         cppcheck.reportError(cond, "style", "It is cleaner to invert the logic.")


@cppcheck.checker
def LambdaAttribute(cfg, data):
    for token in cfg.tokenlist:
        if token.str != ']':
            continue
        if not match(token, "] __device__|__host__ {|("):
            continue
        cppcheck.reportError(
            token, "style",
            "Attributes to lambdas must come after parameter list.")


@cppcheck.checker
def MultipleUnaryOperator(cfg, data):
    for token in cfg.tokenlist:
        if not token.isUnaryOp(token.str):
            continue
        if not token.str in ['+', '-', '~', '!']:
            continue
        if not match(token.astOperand1, "+|-|~|!"):
            continue
        if not token.astOperand1.isUnaryOp(token.astOperand1.str):
            continue
        cppcheck.reportError(token, "style",
                             "Muliple unary operators used together.")


@cppcheck.checker
def MutableVariable(cfg, data):
    for token in cfg.tokenlist:
        if token.str != 'mutable':
            continue
        if not match(token, "mutable %var%"):
            continue
        cppcheck.reportError(token, "style",
                             "Do not create mutable variables.")


@cppcheck.checker
def NestedBlocks(cfg, data):
    for token in cfg.tokenlist:
        if not token.str in ['if', 'while', 'for', 'switch']:
            continue
        block = match(token, "if|while|for|switch (*) { {*}@block }").block
        if not block:
            block = match(token, "; { {*}@block break ; }").block
        if not block:
            continue
        cppcheck.reportError(block, "style", "Block directly inside block.")


@cppcheck.checker
def RedundantCast(cfg, data):
    for token in cfg.tokenlist:
        if not token.variable:
            continue
        m = match(token,
                  "%var%@decl ; %var%@assign = static_cast <*>@cast (*) ;")
        if not m:
            continue
        if m.decl.varId != m.assign.varId:
            continue
        if not simpleMatch(token.previous, "auto"):
            if not isTokensEqual(getVariableDecl(m.decl.variable),
                                 getInnerLink(m.cast),
                                 skip='const|volatile|&|&&|*'):
                continue
        cppcheck.reportError(token, "style", "Static cast is redundant.")


@cppcheck.checker
def RedundantConditionalOperator(cfg, data):
    for token in cfg.tokenlist:
        if token.str != '?':
            continue
        if not match(token, "? true|false : true|false"):
            continue
        cppcheck.reportError(token, "style",
                             "Conditional operator is redundant.")


@cppcheck.checker
def RedundantIfStatement(cfg, data):
    for token in cfg.tokenlist:
        if token.str != 'if':
            continue
        if not match(
                token,
                "if (*) { return true|false ; } else { return true|false ; }"):
            continue
        cppcheck.reportError(token, "style", "The if statement is redundant.")


@cppcheck.checker
def RedundantLocalVariable(cfg, data):
    for token in cfg.tokenlist:
        if not token.variable:
            continue
        m = match(token,
                  "%var%@decl ; %var%@assign = **; return %var%@returned ;")
        if not m:
            continue
        if m.decl.varId != m.assign.varId:
            continue
        if m.decl.varId != m.returned.varId:
            continue
        cppcheck.reportError(
            m.returned, "style",
            "Variable is returned immediately after its declaration, can be simplified to just return expression."
        )


# @cppcheck.checker
# def UnnecessaryElseStatement(cfg, data):
#     for token in cfg.tokenlist:
#         m = match(token, "if (*) {*}@block else@else_statement {")
#         if not m:
#             continue
#         stmt = m.block.link.tokAt(-2)
#         if not match(stmt, "%any% ; }"):
#             continue
#         if not match(stmt, "break|continue"):
#             stmt = stmt.astTop()
#         if not match(stmt, "break|continue|return|throw"):
#             continue
#         cppcheck.reportError(m.else_statement, "style", "Else statement is not necessary.")


@cppcheck.checker
def UnnecessaryEmptyCondition(cfg, data):
    for token in cfg.tokenlist:
        if token.str != 'if':
            continue
        m = match(token, "if (*)@if_cond { for (*)@for_cond {*} }")
        if not m:
            continue
        cond = m.if_cond.astOperand2
        if match(cond, "!"):
            cond = cond.astOperand1
        if not match(cond.tokAt(-2), ". empty ("):
            continue
        container = cond.tokAt(-2).astOperand1
        if not container.varId:
            continue
        if not match(m.for_cond.astOperand2, ":"):
            continue
        container_iter = m.for_cond.astOperand2.astOperand2
        if container_iter.varId != container.varId:
            continue
        cppcheck.reportError(
            container, "style",
            "Unnecessary check for empty before for range loop.")


@cppcheck.checker
def UseDeviceLaunch(cfg, data):
    for token in cfg.tokenlist:
        if not simpleMatch(token, "hipLaunchKernelGGL ("):
            continue
        cppcheck.reportError(token, "style", "Use device::launch instead.")


@cppcheck.checker
def UseManagePointer(cfg, data):
    functions = {
        "fclose", "free", "hipFree", "hipHostFree", "hipFreeArray",
        "hipMemFree", "hipStreamDestroy", "hipEventDestroy", "hipArrayDestroy",
        "hipCtxDestroy", "hipDestroyTextureObject", "hipDestroySurfaceObject",
        "miirDestroyHandle"
    }
    for token in cfg.tokenlist:
        if not isFunctionCall(token):
            continue
        if not token.str in functions:
            continue
        cppcheck.reportError(token, "style",
                             "Use manage pointer for resource management.")


@cppcheck.checker
def UseSmartPointer(cfg, data):
    for token in cfg.tokenlist:
        if token.str != 'new':
            continue
        if not match(token, "new %name%"):
            continue
        cppcheck.reportError(token, "style",
                             "Use manage pointer for resource management.")


@cppcheck.checker
def useStlAlgorithms(cfg, data):
    copy_functions = {"memcpy", "strcpy", "strncpy", "strcat", "strncat"}
    for token in cfg.tokenlist:
        if not isFunctionCall(token):
            continue
        if token.str in copy_functions:
            cppcheck.reportError(token, "style", "Use std::copy instead.")
        elif token.str == 'memset':
            cppcheck.reportError(token, "style", "Use std::fill instead.")
        elif token.str == 'memcmp':
            cppcheck.reportError(token, "style",
                                 "Use std::equal_range instead.")
        elif token.str == 'memchr':
            cppcheck.reportError(token, "style", "Use std::find instead.")


@cppcheck.checker
def MatcherNestedParentheses(cfg, data):
    for token in cfg.tokenlist:
        if not simpleMatch(token, "matcher ( ) const {"):
            continue
        for tok2 in token.tokAt(4).forward(token.linkAt(4)):
            if not simpleMatch(tok2, ") ) ) )"):
                continue
            if simpleMatch(tok2.link.previous, "bind"):
                continue
            cppcheck.reportError(
                tok2, "style",
                "Too many nested parentheses can affect readability; consider using variables instead."
            )
            break



#####################################################################################
# The MIT License (MIT)
#
# Copyright (c) 2015-2023 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#####################################################################################
import migraphx


def test_conv_relu():
    p = migraphx.parse_onnx("conv_relu_maxpool_test.onnx")
    print(p)
    print("Compiling ...")
    # set offload_copy, fast_match to true
    p.compile(migraphx.get_target("gpu"), True, True)
    print(p)
    params = {}

    for key, value in p.get_parameter_shapes().items():
        print("Parameter {} -> {}".format(key, value))
        params[key] = migraphx.generate_argument(value)

    r = p.run(params)
    print(r)


def test_sub_uint64():
    p = migraphx.parse_onnx("implicit_sub_bcast_test.onnx")
    print(p)
    print("Compiling ...")
    p.compile(migraphx.get_target("gpu"))
    print(p)
    params = {}

    shapes = p.get_parameter_shapes()
    params["0"] = migraphx.create_argument(
        migraphx.shape(type='uint64_type', lens=shapes["0"].lens()),
        list(range(120)))
    params["1"] = migraphx.create_argument(
        migraphx.shape(type='uint64_type', lens=shapes["1"].lens()),
        list(range(20)))

    r = p.run(params)
    print(r)


def test_neg_int64():
    p = migraphx.parse_onnx("neg_test.onnx")
    print(p)
    print("Compiling ...")
    p.compile(migraphx.get_target("gpu"))
    print(p)
    params = {}

    shapes = p.get_parameter_shapes()
    params["0"] = migraphx.create_argument(
        migraphx.shape(type='int64_type', lens=shapes["0"].lens()),
        list(range(6)))

    r = p.run(params)
    print(r)


def test_nonzero():
    p = migraphx.parse_onnx("nonzero_dynamic_test.onnx")
    print(p)
    print("Compiling ...")
    p.compile(migraphx.get_target("gpu"))
    print(p)
    params = {}

    shapes = p.get_parameter_shapes()
    params["data"] = migraphx.create_argument(
        migraphx.shape(type='bool_type', lens=shapes["data"].lens()),
        [1, 1, 0, 1])

    r = p.run(params)
    print(r)


def test_fp16_imagescaler():
    p = migraphx.parse_onnx("imagescaler_half_test.onnx")
    print(p)
    s1 = p.get_output_shapes()[-1]
    print("Compiling ...")
    p.compile(migraphx.get_target("gpu"))
    print(p)
    s2 = p.get_output_shapes()[-1]
    assert s1 == s2

    params = {}
    shapes = p.get_parameter_shapes()
    params["0"] = migraphx.generate_argument(
        migraphx.shape(type='half_type', lens=shapes["0"].lens()), 768)

    r = p.run(params)[-1]
    print(r)


def test_if_pl():
    p = migraphx.parse_onnx("if_pl_test.onnx")
    print(p)
    s1 = p.get_output_shapes()[-1]
    print("Compiling ...")
    p.compile(migraphx.get_target("gpu"))
    print(p)
    s2 = p.get_output_shapes()[-1]
    assert s1 == s2

    params = {}
    shapes = p.get_parameter_shapes()
    params["x"] = migraphx.fill_argument(
        migraphx.shape(type='float_type', lens=shapes["x"].lens()), 1)
    params["y"] = migraphx.fill_argument(
        migraphx.shape(type='float_type', lens=shapes["y"].lens()), 2.0)
    params["cond"] = migraphx.fill_argument(
        migraphx.shape(type="bool", lens=[1], strides=[0]), 1)

    r = p.run(params)[-1]
    print(r)


def test_dyn_batch():
    a = migraphx.shape.dynamic_dimension(1, 4, {2, 4})
    b = migraphx.shape.dynamic_dimension(3, 3)
    c = migraphx.shape.dynamic_dimension(32, 32)
    dd_map = {"0": [a, b, c, c]}
    p = migraphx.parse_onnx("conv_relu_maxpool_test.onnx",
                            map_dyn_input_dims=dd_map)
    print(p)
    print("Compiling ...")
    p.compile(migraphx.get_target("gpu"))
    print(p)

    def run_prog(batch_size):
        params = {}
        for key, value in p.get_parameter_shapes().items():
            # convert to a static shape
            if value.dynamic():
                dds = value.dyn_dims()
                new_lens = []
                for dd in dds:
                    if dd.is_fixed():
                        new_lens.append(dd.min)
                    else:
                        new_lens.append(batch_size)
                s = migraphx.shape(type=value.type_string(), lens=new_lens)
            else:
                s = value
            print("Parameter {} -> {}".format(key, s))
            params[key] = migraphx.generate_argument(s)
        r = p.run(params)
        print(r)

    run_prog(1)
    run_prog(2)
    run_prog(3)
    run_prog(4)


test_conv_relu()
test_sub_uint64()
test_neg_int64()
test_fp16_imagescaler()
test_if_pl()
test_nonzero()
test_dyn_batch()



#####################################################################################
# The MIT License (MIT)
#
# Copyright (c) 2015-2024 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#####################################################################################
import cv2
import numpy as np


def process_image(pil_image, shape=(224, 224), means=None):
    # OpenCV default is BGR, the image is in RGB
    image = cv2.cvtColor(np.array(pil_image), cv2.COLOR_RGB2BGR)
    # But we need it to be in RGB, not sure how to import and not convert
    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
    w, h = shape
    image = resize_with_aspectratio(image, h, w, inter_pol=cv2.INTER_AREA)
    image = center_crop(image, h, w)
    image = np.asarray(image, dtype="float32")
    # Normalize image.
    if means:
        means = np.array(means, dtype=np.float32)
        image -= means
    # RGB is channel_last (HWC), transpose to channel_first (CHW)
    image = image.transpose([2, 0, 1])
    image = image[np.newaxis, :]
    return image


def center_crop(img, out_height, out_width):
    height, width, _ = img.shape
    left = int((width - out_width) / 2)
    right = int((width + out_width) / 2)
    top = int((height - out_height) / 2)
    bottom = int((height + out_height) / 2)
    img = img[top:bottom, left:right]
    return img


def resize_with_aspectratio(img,
                            out_height,
                            out_width,
                            scale=87.5,
                            inter_pol=cv2.INTER_LINEAR):
    height, width, _ = img.shape
    new_height = int(100.0 * out_height / scale)
    new_width = int(100.0 * out_width / scale)
    if height > width:
        w = new_width
        h = int(new_height * height / width)
    else:
        h = new_height
        w = int(new_width * width / height)
    img = cv2.resize(img, (w, h), interpolation=inter_pol)
    return img



#####################################################################################
# The MIT License (MIT)
#
# Copyright (c) 2015-2023 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#####################################################################################
import migraphx


def test_instruction_shape():
    p = migraphx.program()
    mm = p.get_main_module()
    input_shape = migraphx.shape(lens=[4, 4, 64], type="half_type")
    i = mm.add_parameter("x", input_shape)
    i2 = mm.add_instruction(migraphx.op("reshape", dims=[16, 64]), [i])
    out_shape = i2.shape()

    assert out_shape.lens() == [16, 64]
    assert out_shape.strides() == [64, 1]
    assert out_shape.type_string() == "half_type"


def test_instruction_op():
    p = migraphx.program()
    mm = p.get_main_module()
    input_shape = migraphx.shape(lens=[2, 24])
    i = mm.add_parameter("x", input_shape)
    i2 = mm.add_instruction(migraphx.op("relu"), [i])
    out_op = i2.op()

    assert out_op.name() == "relu"


if __name__ == "__main__":
    test_instruction_shape()
    test_instruction_op()



#####################################################################################
# The MIT License (MIT)
#
# Copyright (c) 2015-2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#####################################################################################
import migraphx
import onnx
from onnx import helper, TensorProto, checker
import numpy as np
import os
import argparse


# Utility function to map MIGraphX types to ONNX data types
def get_dtype(instruction):
    type_mapping = {
        'float_type': TensorProto.FLOAT,
        'bf16_type': TensorProto.BFLOAT16,
        'half_type': TensorProto.FLOAT16
    }
    return type_mapping[instruction.shape().type_string()]


# Utility function to get the shape of an instruction
def get_shape(instruction):
    if isinstance(instruction, list):
        raise ValueError("Expected instruction, got a list.")
    return instruction.shape().lens()


# Utility function to map MIGraphX operations to ONNX operations
def map_operation(operation):
    mxr_to_onnx_op = {
        "dot": "MatMul",
        "mul": "Mul",
        "add": "Add",
        "multibroadcast": "Expand",
        "erf": "Erf",
        "tanh": "Tanh",
        "exp": "Exp",
        "div": "Div",
        "relu": "Relu",
        "pow": "Pow"
    }

    if operation not in mxr_to_onnx_op:
        raise NotImplementedError(f"Operation '{operation}' is not supported.")
    return mxr_to_onnx_op[operation]


# Helper function to create ONNX nodes for specific operations
def create_node(instruction, parameters, node_name, n, initializers):
    if node_name == "multibroadcast" or node_name == "reshape":
        shape_key = "out_lens" if node_name == "multibroadcast" else "dims"
        shape_array = np.array(parameters[shape_key], dtype=np.int64)
        initializer_name = f"{node_name}_shape_{n}"

        initializers.append(
            helper.make_tensor(name=initializer_name,
                               data_type=TensorProto.INT64,
                               dims=shape_array.shape,
                               vals=shape_array.flatten().tolist()))
        return helper.make_node(
            map_operation(node_name),
            inputs=[str(hash(i))
                    for i in instruction.inputs()] + [initializer_name],
            outputs=[str(hash(instruction))])

    elif node_name == "transpose":
        return helper.make_node(
            "Transpose",
            inputs=[str(hash(i)) for i in instruction.inputs()],
            outputs=[str(hash(instruction))],
            perm=parameters["permutation"])

    elif node_name == "convolution":
        return helper.make_node(
            "Conv",
            inputs=[str(hash(i)) for i in instruction.inputs()],
            outputs=[str(hash(instruction))
                     ],  #[str(hash(i)) for i in instruction.outputs()],
            dilations=parameters["dilation"],
            group=parameters["group"],
            pads=parameters["padding"],
            strides=parameters["stride"])

    return helper.make_node(
        map_operation(node_name),
        inputs=[str(hash(i)) for i in instruction.inputs()],
        outputs=[str(hash(instruction))])


# Main function to convert MIGraphX module to ONNX model
def generate_onnx(module):
    inputs = {}
    operations = []
    initializers = []
    n = 0  # Node counter
    output = None

    for instruction in module:
        op_name = instruction.op().name()

        # Handle input nodes
        if op_name in ["@literal", "@param"]:

            inputs[str(hash(instruction))] = helper.make_tensor_value_info(
                str(hash(instruction)), get_dtype(instruction),
                get_shape(instruction))

        # Handle computational nodes
        elif "@" not in op_name:
            n += 1
            parameters = instruction.op().values()

            operations.append(
                create_node(instruction, parameters, op_name, n, initializers))

        # Handle return node
        elif op_name == "@return":

            output = [
                helper.make_tensor_value_info(str(hash(i)), get_dtype(i),
                                              get_shape(i))
                for i in instruction.inputs()
            ]

    # Create the ONNX graph
    graph = helper.make_graph(nodes=operations,
                              name="Graph",
                              inputs=list(inputs.values()),
                              initializer=initializers,
                              outputs=output if output else [])

    return helper.make_model(graph, producer_name="onnx-dot-add-example")


# Main function to process MIGraphX files and generate ONNX models
def main(mxr_directory_path, onnx_directory_path):
    for file_name in os.listdir(mxr_directory_path):
        file_path = os.path.join(mxr_directory_path, file_name)
        if ".mxr" in file_path:
            try:
                program = migraphx.load(file_path)
                module = program.get_main_module()
                model = generate_onnx(module)

                # Validate the generated ONNX model
                try:
                    checker.check_model(model)
                    print(f"ONNX model for {file_path} is valid.")
                except onnx.checker.ValidationError as e:
                    print(f"Validation failed for {file_path}: {e}")
                except Exception as e:
                    print(
                        f"Unexpected error during validation for {file_path}: {e}"
                    )

                os.makedirs(onnx_directory_path, exist_ok=True)
                onnx_file_path = os.path.join(onnx_directory_path,
                                              file_name.replace("mxr", "onnx"))
                onnx.save(model, onnx_file_path)

            except Exception as e:
                print(f"Error processing {file_path}: {e}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Process MXR files and generate ONNX models.")
    parser.add_argument("mxr_directory_path",
                        type=str,
                        help="Path to the directory containing MXR files.")
    parser.add_argument(
        "onnx_directory_path",
        type=str,
        help="Path to the directory where ONNX models will be saved.")

    args = parser.parse_args()
    main(args.mxr_directory_path, args.onnx_directory_path)



import torch
import os


def make_directories(dir_path, model_name):
    if not os.path.exists(f"{dir_path}/{model_name}/"):
        os.makedirs(f"{dir_path}/{model_name}/")


def export_rnnt_onnx(model, x, dir_path):

    for component in ['rnnt_encoder', 'rnnt_prediction', 'rnnt_joint']:
        make_directories(dir_path, component)

    seq_length, batch_size, feature_length = x.shape
    inp = torch.randn([seq_length, batch_size, feature_length]).to('cuda')
    feature_length = torch.LongTensor([seq_length]).to('cuda')
    x_padded, x_lens = model.encoder(inp, feature_length)
    torch.onnx.export(model.encoder, (inp, feature_length),
                      f"{dir_path}/rnnt_encoder/model.onnx",
                      opset_version=12,
                      input_names=['input', 'feature_length'],
                      output_names=['x_padded', 'x_lens'],
                      dynamic_axes={'input': {
                          0: 'seq_len',
                          1: 'batch'
                      }})

    symbol = torch.LongTensor([[20]]).to('cuda')
    hidden = torch.randn([2, batch_size,
                          320]).to('cuda'), torch.randn([2, batch_size,
                                                         320]).to('cuda')
    g, hidden = model.prediction.forward(symbol, hidden)
    torch.onnx.export(model.prediction, (symbol, hidden),
                      f"{dir_path}/rnnt_prediction/model.onnx",
                      opset_version=12,
                      input_names=['symbol', 'hidden_in_1', 'hidden_in_2'],
                      output_names=['g', 'hidden_out_1', 'hidden_out_2'],
                      dynamic_axes={
                          'symbol': {
                              0: 'batch'
                          },
                          'hidden_in_1': {
                              1: 'batch'
                          },
                          'hidden_in_2': {
                              1: 'batch'
                          }
                      })

    f = torch.randn([batch_size, 1, 1024]).to('cuda')
    model.joint.forward(f, g).to('cuda')
    torch.onnx.export(model.joint, (f, g),
                      f"{dir_path}/rnnt_joint/model.onnx",
                      opset_version=12,
                      input_names=['0', '1'],
                      output_names=['result'],
                      dynamic_axes={
                          '0': {
                              0: 'batch'
                          },
                          '1': {
                              0: 'batch'
                          }
                      })



#####################################################################################
# The MIT License (MIT)
#
# Copyright (c) 2015-2024 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#####################################################################################

from argparse import ArgumentParser
from transformers import WhisperProcessor
from datasets import load_dataset
from pydub import AudioSegment

import migraphx as mgx
import os
import numpy as np
import sys
import time
from functools import wraps


# measurement helper
def measure(fn):
    @wraps(fn)
    def measure_ms(*args, **kwargs):
        start_time = time.perf_counter_ns()
        result = fn(*args, **kwargs)
        end_time = time.perf_counter_ns()
        print(
            f"Elapsed time for {fn.__name__}: {(end_time - start_time) * 1e-6:.4f} ms\n"
        )
        return result

    return measure_ms


def get_args():
    parser = ArgumentParser()
    parser.add_argument(
        "-a",
        "--audio",
        type=str,
        help="Path to audio file. Default: HF test dataset",
    )

    parser.add_argument(
        "-l",
        "--log-process",
        action="store_true",
        help="Print the current state of transcribing.",
    )

    return parser.parse_args()


class WhisperMGX():
    def __init__(self):
        model_id = "openai/whisper-tiny.en"
        print(f"Using {model_id}")

        print("Creating Whisper processor")
        self.processor = WhisperProcessor.from_pretrained(model_id)

        self.decoder_start_token_id = 50257  # <|startoftranscript|>
        self.eos_token_id = 50256  # "<|endoftext|>"
        self.notimestamps = 50362  # <|notimestamps|>
        self.max_length = 448
        self.sot = [self.decoder_start_token_id, self.notimestamps]

        print("Load models...")
        self.encoder_model = WhisperMGX.load_mgx_model(
            "encoder", {"input_features": [1, 80, 3000]})
        self.decoder_model = WhisperMGX.load_mgx_model(
            "decoder", {
                "input_ids": [1, self.max_length],
                "attention_mask": [1, self.max_length],
                "encoder_hidden_states": [1, 1500, 384]
            })

    @staticmethod
    @measure
    def load_audio_from_file(filepath):
        audio = AudioSegment.from_file(filepath)
        # Only 16k is supported
        audio = audio.set_frame_rate(16000)
        data = np.array(audio.get_array_of_samples(), dtype=np.float32)
        data /= np.max(np.abs(data))
        return data, audio.frame_rate

    @staticmethod
    @measure
    def load_mgx_model(name, shapes):
        file = f"models/whisper-tiny.en_modified/{name}_model"
        print(f"Loading {name} model from {file}")
        if os.path.isfile(f"{file}.mxr"):
            print("Found mxr, loading it...")
            model = mgx.load(f"{file}.mxr", format="msgpack")
        elif os.path.isfile(f"{file}.onnx"):
            print("Parsing from onnx file...")
            model = mgx.parse_onnx(f"{file}.onnx", map_input_dims=shapes)
            model.compile(mgx.get_target("gpu"))
            print(f"Saving {name} model to mxr file...")
            mgx.save(model, f"{file}.mxr", format="msgpack")
        else:
            print(f"No {name} model found. Please download it and re-try.")
            sys.exit(1)
        return model

    @property
    def initial_decoder_inputs(self):
        input_ids = np.array([
            self.sot + [self.eos_token_id] * (self.max_length - len(self.sot))
        ])
        # 0 masked | 1 un-masked
        attention_mask = np.array([[1] * len(self.sot) + [0] *
                                   (self.max_length - len(self.sot))])
        return (input_ids, attention_mask)

    @measure
    def get_input_features_from_sample(self, sample_data, sampling_rate):
        return self.processor(sample_data,
                              sampling_rate=sampling_rate,
                              return_tensors="np").input_features

    @measure
    def encode_features(self, input_features):
        return np.array(
            self.encoder_model.run(
                {"input_features": input_features.astype(np.float32)})[0])

    def decode_step(self, input_ids, attention_mask, hidden_states):
        return np.array(
            self.decoder_model.run({
                "input_ids":
                input_ids.astype(np.int64),
                "attention_mask":
                attention_mask.astype(np.int64),
                "encoder_hidden_states":
                hidden_states.astype(np.float32)
            })[0])

    @measure
    def generate(self, input_features, log_process=False):
        hidden_states = self.encode_features(input_features)
        input_ids, attention_mask = self.initial_decoder_inputs
        for timestep in range(len(self.sot) - 1, self.max_length):
            # get logits for the current timestep
            logits = self.decode_step(input_ids, attention_mask, hidden_states)
            # greedily get the highest probable token
            new_token = np.argmax(logits[0][timestep])

            # add it to the tokens and unmask it
            input_ids[0][timestep + 1] = new_token
            attention_mask[0][timestep + 1] = 1

            if log_process:
                print("Transcribing: " + ''.join(
                    self.processor.decode(input_ids[0][:timestep + 1],
                                          skip_special_tokens=True)),
                      end='\r')

            if new_token == self.eos_token_id:
                break

        if log_process:
            print(flush=True)

        return ''.join(
            self.processor.decode(input_ids[0][:timestep + 1],
                                  skip_special_tokens=True))


if __name__ == "__main__":
    args = get_args()

    if args.audio:
        data, fr = WhisperMGX.load_audio_from_file(args.audio)
        ds = [{"audio": {"array": data, "sampling_rate": fr}}]
    else:
        # load dummy dataset and read audio files
        ds = load_dataset("hf-internal-testing/librispeech_asr_dummy",
                          "clean",
                          split="validation")

    w = WhisperMGX()

    for idx, data in enumerate(ds):
        print(f"#{idx+1}/{len(ds)} Sample...")
        sample = data["audio"]
        input_features = w.get_input_features_from_sample(
            sample["array"], sample["sampling_rate"])
        result = w.generate(input_features, log_process=args.log_process)
        print(f"Result: {result}")



#####################################################################################
#  The MIT License (MIT)
#
#  Copyright (c) 2015-2025 Advanced Micro Devices, Inc. All rights reserved.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.
#####################################################################################
import subprocess
import shlex
import datetime


def run(cmd, verbose=False, **kwargs):
    """
    Run a shell command with optional output capture and verbosity.

    Args:
        cmd (str | list): The command to run.
        cwd (str): The current working directory.
        verbose (bool): If True, print the command being run.
        **kwargs: Extra arguments passed to subprocess.run().

    Returns:
        str | None: If capture is True, returns the command's stdout as a string; otherwise None.
    """
    if verbose:
        print(cmd if isinstance(cmd, str) else shlex.join(cmd))

    result = subprocess.run(cmd,
                            shell=isinstance(cmd, str),
                            check=True,
                            capture_output=True,
                            text=True,
                            **kwargs).stdout.strip()

    if verbose:
        print(result)

    return result


def get_top():
    """
    Return the absolute path to the root directory of the Git repository.
    Useful as the base path for running Git commands or locating files.
    """
    return run("git rev-parse --show-toplevel")


def get_merge_base(branch):
    """
    Get the common ancestor (merge base) commit hash between the current branch
    and the given branch. This is the commit where they diverged, used to compare changes.
    """
    head = run("git rev-parse --abbrev-ref HEAD")  # Get current branch name
    return run(f"git merge-base {branch} {head}"
               )  # Get merge base SHA between current and target branch


def get_changed_files(branch):
    """
    Return a list of files that are staged for commit and have changed
    compared to the merge base with the given branch.
    Only includes modified or added files (excludes deleted files).
    """
    base = get_merge_base(branch)
    return run(f"git diff --cached --name-only --diff-filter=d {base}",
               cwd=get_top()).splitlines()


def get_all_files():
    """
    Return a list of all tracked files in the Git repository.
    """
    return run("git ls-files", cwd=get_top()).splitlines()


def get_latest_commit_year(file):
    """
    Get year of latest commit given file 
    """
    date_str = run(f"git log -1 --format=%cd --date=short {file}",
                   cwd=get_top())
    return datetime.datetime.strptime(date_str, '%Y-%m-%d').year



#####################################################################################
# The MIT License (MIT)
#
# Copyright (c) 2015-2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#####################################################################################
import argparse
import numpy as np
import onnxruntime as ort
import sys


def parse_args():
    parser = argparse.ArgumentParser(
        description=
        'ONNX Runtime python driver. Use to run onnx models using ONNX Runtime'
    )
    file_args = parser.add_argument_group(title='file type arguments')
    file_args.add_argument('--onnx', type=str, help='path to onnx file')
    parser.add_argument('--provider',
                        type=str,
                        default='CPUExecutionProvider',
                        help='execution provider for onnx runtime \
                                (default = CPUExecutionProvider)')
    parser.add_argument('--fill1',
                        action='store_true',
                        help='fill all arguments with a value of 1')
    parser.add_argument('--fill0',
                        action='store_true',
                        help='fill all arguments with a value of 0')
    parser.add_argument(
        '--default-dim-value',
        type=int,
        default=1,
        help='default dim value (if any specified in onnx file)')
    parser.add_argument('--verbose',
                        action='store_true',
                        help='show verbose information (for debugging)')
    parser.add_argument('--input-dim',
                        type=str,
                        action='append',
                        help='specify input parameter dimension \
                                with the following format --input-dim input_name:dim0,dim1,dim2...'
                        )

    parser.add_argument('--ort-logging',
                        dest="ort_logging",
                        action='store_true',
                        default=False,
                        help='Turn on ort VERBOSE logging via session options')

    args = parser.parse_args()

    return args, parser


def get_np_datatype(in_type):
    datatypes = {
        'tensor(double)': np.float64,
        'tensor(float)': np.float32,
        'tensor(float16)': np.half,
        'tensor(int64)': np.int64,
        'tensor(uint64)': np.uint64,
        'tensor(int32)': np.int32,
        'tensor(uint32)': np.uint32,
        'tensor(int16)': np.int16,
        'tensor(uint16)': np.uint16,
        'tensor(int8)': np.int8,
        'tensor(uint8)': np.uint8,
    }
    return datatypes[in_type]


def main():
    args, parser = parse_args()

    if args.onnx == None:
        print('Error: please specify an onnx file')
        parser.print_help()
        sys.exit(-1)

    model_name = args.onnx

    custom_inputs = args.input_dim

    input_dims = {}
    if custom_inputs != None:
        for input in custom_inputs:
            input_dim = ''.join(input.split(':')[:-1])
            dims = [int(dim) for dim in input.split(':')[-1].split(',')]
            input_dims[input_dim] = dims

    sess_op = ort.SessionOptions()

    if args.ort_logging:
        sess_op.log_verbosity_level = 0
        sess_op.log_severity_level = 0

    sess = ort.InferenceSession(model_name,
                                sess_options=sess_op,
                                providers=[args.provider])

    test_inputs = {}
    for input in sess.get_inputs():
        name = input.name
        in_shape = input.shape
        in_type = input.type

        for idx, dim in enumerate(in_shape):
            if not isinstance(dim, int):
                if args.verbose:
                    print(
                        f'''Dim param found at index {idx}: {dim}. '''
                        f'''Setting to default dim value of {args.default_dim_value}.'''
                    )
                in_shape[idx] = args.default_dim_value

        if name in input_dims:
            in_shape = input_dims[name]

        if args.verbose:
            print(f'Parameter {name} -> {in_shape}, {in_type}')

        if not args.fill1 and not args.fill0:
            test_input = np.random.rand(*(in_shape)).astype(
                get_np_datatype(in_type))
        elif not args.fill0:
            test_input = np.ones(in_shape).astype(get_np_datatype(in_type))
        else:
            test_input = np.zeros(in_shape).astype(get_np_datatype(in_type))
        test_inputs[name] = test_input

    ort_params = {}
    for input in sess.get_inputs():
        ort_params[input.name] = test_inputs[input.name]

    try:
        pred_ort = sess.run(None, ort_params)
        if args.verbose:
            for idx, output in enumerate(pred_ort):
                print(f'output at index {idx}: {output}')
    except Exception as e:
        if any(input_dims):
            print(
                'Error: custom input dim may not be compatible with onnx runtime'
            )
        raise e
    print('onnx runtime driver completed successfully.')


if __name__ == '__main__':
    main()



#####################################################################################
# The MIT License (MIT)
#
# Copyright (c) 2015-2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#####################################################################################
import api


def bad_param_error(msg):
    return 'MIGRAPHX_THROW(migraphx_status_bad_param, "{}")'.format(msg)


api.error_type = 'migraphx_status'
api.success_type = 'migraphx_status_success'
api.try_wrap = 'migraphx::try_'
api.bad_param_error = bad_param_error


@api.cwrap('migraphx::shape::type_t')
def shape_type_wrap(p):
    if p.returns:
        p.add_param('migraphx_shape_datatype_t *')
        p.bad_param('${name} == nullptr', 'Null pointer')
        p.write = ['*${name} = migraphx::to_shape_type(${result})']
    else:
        p.add_param('migraphx_shape_datatype_t')
        p.read = 'migraphx::to_shape_type(${name})'


def auto_handle(*args, **kwargs):
    def with_handle(f):
        return api.handle('migraphx_' + f.__name__, 'migraphx::' + f.__name__,
                          *args, **kwargs)(f)

    return with_handle


@api.handle('migraphx_optimals', 'std::set<size_t>')
def optimals(h):
    h.constructor('create',
                  api.params(ptr='const size_t*', size='size_t'),
                  fname='migraphx::make_set<size_t>')


@api.handle('migraphx_dynamic_dimension', 'migraphx::shape::dynamic_dimension')
def dynamic_dimension(h):
    h.constructor('create_min_max', api.params(min='size_t', max='size_t'))
    h.constructor(
        'create_min_max_optimals',
        api.params(min='size_t', max='size_t', optimals='std::set<size_t>'))
    h.method('is_fixed', returns='bool', const=True)
    h.method('equal',
             api.params(x='const migraphx::shape::dynamic_dimension&'),
             invoke='migraphx::equal($@)',
             returns='bool',
             const=True)


@api.handle('migraphx_dynamic_dimensions',
            'std::vector<migraphx::shape::dynamic_dimension>')
def dynamic_dimensions(h):
    h.constructor(
        'create',
        api.params(ptr='const const_migraphx_dynamic_dimension_t*',
                   size='size_t'),
        fname='migraphx::to_obj_vector<const_migraphx_dynamic_dimension_t>')
    h.method('size', returns='size_t')
    h.method('get',
             api.params(idx='size_t'),
             fname='at',
             cpp_name='operator[]',
             returns='const migraphx::shape::dynamic_dimension&')


@auto_handle()
def shape(h):
    h.constructor(
        'create',
        api.params(type='migraphx::shape::type_t',
                   lengths='std::vector<size_t>'))
    h.constructor(
        'create_with_strides',
        api.params(type='migraphx::shape::type_t',
                   lengths='std::vector<size_t>',
                   strides='std::vector<size_t>'))
    h.constructor('create_scalar', api.params(type='migraphx::shape::type_t'))
    h.constructor(
        'create_dynamic',
        api.params(type='migraphx::shape::type_t',
                   dims='std::vector<migraphx::shape::dynamic_dimension>'))
    h.method('lengths',
             fname='lens',
             returns='const std::vector<size_t>&',
             const=True)
    h.method('strides', returns='const std::vector<size_t>&', const=True)
    h.method('dyn_dims',
             returns='std::vector<migraphx::shape::dynamic_dimension>',
             const=True)
    h.method('type', returns='migraphx::shape::type_t', const=True)
    h.method('elements', returns='size_t', const=True)
    h.method('bytes', returns='size_t', const=True)
    h.method('ndim', returns='size_t', const=True)
    h.method('equal',
             api.params(x='const migraphx::shape&'),
             invoke='migraphx::equal($@)',
             returns='bool',
             const=True)
    h.method('standard', returns='bool', const=True)
    h.method('dynamic', returns='bool', const=True)
    h.method('index', api.params(i='size_t'), returns='size_t', const=True)


@auto_handle()
def argument(h):
    h.constructor('create',
                  api.params(shape='const migraphx::shape&', buffer='void*'))
    h.constructor('create_empty', api.params(shape='const migraphx::shape&'))
    h.method('shape',
             fname='get_shape',
             cpp_name='get_shape',
             returns='const migraphx::shape&',
             const=True)
    h.method('buffer',
             fname='data',
             cpp_name='data',
             returns='char*',
             const=True)
    h.method('equal',
             api.params(x='const migraphx::argument&'),
             invoke='migraphx::equal($@)',
             returns='bool',
             const=True)


api.add_function('migraphx_argument_save',
                 api.params(a='const migraphx::argument&', filename='const char*'),
                 fname='migraphx::save_argument'
                 )

api.add_function('migraphx_argument_load',
                 api.params(filename='const char*'),
                 fname='migraphx::load_argument',
                 returns='migraphx::argument'
                 )

api.add_function('migraphx_argument_generate',
                 api.params(s='const migraphx::shape&', seed='size_t'),
                 fname='migraphx::generate_argument',
                 returns='migraphx::argument')


@auto_handle()
def target(h):
    h.constructor('create',
                  api.params(name='const char*'),
                  fname='migraphx::get_target')


@api.handle('migraphx_program_parameter_shapes',
            'std::unordered_map<std::string, migraphx::shape>')
def program_parameter_shapes(h):
    h.method('size', returns='size_t')
    h.method('get',
             api.params(name='const char*'),
             fname='at',
             cpp_name='operator[]',
             returns='const migraphx::shape&')
    h.method('names',
             invoke='migraphx::get_names(${program_parameter_shapes})',
             returns='std::vector<const char*>')


@api.handle('migraphx_program_parameters',
            'std::unordered_map<std::string, migraphx::argument>')
def program_parameters(h):
    h.constructor('create')
    h.method('add',
             api.params(name='const char*',
                        argument='const migraphx::argument&'),
             invoke='${program_parameters}[${name}] = ${argument}')


@api.handle('migraphx_arguments', 'std::vector<migraphx::argument>')
def arguments(h):
    h.method('size', returns='size_t')
    h.method('get',
             api.params(idx='size_t'),
             fname='at',
             cpp_name='operator[]',
             returns='const migraphx::argument&')


@api.handle('migraphx_shapes', 'std::vector<migraphx::shape>')
def shapes(h):
    h.method('size', returns='size_t')
    h.method('get',
             api.params(idx='size_t'),
             fname='at',
             cpp_name='operator[]',
             returns='const migraphx::shape&')


@api.handle('migraphx_instruction', 'migraphx::instruction_ref')
def instruction(h):
    pass


@api.handle('migraphx_instructions', 'std::vector<migraphx::instruction_ref>')
def instructions(h):
    h.constructor(
        'create',
        api.params(ptr='const const_migraphx_instruction_t*', size='size_t'),
        fname='migraphx::to_obj_vector<const_migraphx_instruction_t>')


@api.handle('migraphx_modules', 'std::vector<migraphx::module*>')
def modules(h):
    h.constructor('create',
                  api.params(ptr='migraphx_module_t*', size='size_t'),
                  fname='migraphx::to_objptr_vector<migraphx::module*>')


@auto_handle(ref=True)
def module(h):
    h.constructor('create', api.params(name='std::string'))
    h.method('print', invoke='migraphx::print_module($@)', const=True)
    h.method('add_instruction',
             api.params(op='migraphx::operation',
                        args='std::vector<migraphx::instruction_ref>'),
             returns='migraphx::instruction_ref')
    h.method('add_instruction_with_mod_args',
             api.params(op='migraphx::operation',
                        args='std::vector<migraphx::instruction_ref>',
                        module_refs='std::vector<migraphx::module*>'),
             fname='add_instruction',
             returns='migraphx::instruction_ref')
    h.method('add_literal',
             api.params(shape='const migraphx::shape&', buffer='const char*'),
             returns='migraphx::instruction_ref')
    h.method('add_parameter',
             api.params(name='const char*', shape='const migraphx::shape&'),
             returns='migraphx::instruction_ref')
    h.method('add_return',
             api.params(args='std::vector<migraphx::instruction_ref>'),
             returns='migraphx::instruction_ref')
    h.method('add_allocation',
             api.params(s='const migraphx::shape&'),
             invoke='migraphx::add_allocation($@)',
             returns='migraphx::instruction_ref')


@auto_handle()
def program(h):
    h.constructor('create')
    h.method('get_main_module', returns='migraphx::module*')
    h.method('create_module',
             api.params(name='const char*'),
             returns='migraphx::module*')
    h.method(
        'compile',
        api.params(target='migraphx::target',
                   options='migraphx::compile_options'))
    h.method('get_parameter_shapes',
             returns='std::unordered_map<std::string, migraphx::shape>')
    h.method('get_output_shapes',
             invoke='migraphx::get_output_shapes($@)',
             returns='std::vector<migraphx::shape>')
    h.method('print', invoke='migraphx::print_program($@)', const=True)
    h.method('sort')
    h.method('run',
             api.params(
                 params='std::unordered_map<std::string, migraphx::argument>'),
             invoke='migraphx::run($@)',
             returns='std::vector<migraphx::argument>')
    h.method('run_async',
             api.params(
                 params='std::unordered_map<std::string, migraphx::argument>',
                 s='void*',
                 name='const char *'),
             invoke='migraphx::run_async($@)',
             returns='std::vector<migraphx::argument>')
    h.method('equal',
             api.params(x='const migraphx::program&'),
             invoke='migraphx::equal($@)',
             returns='bool',
             const=True)
    h.method('experimental_get_context',
             invoke='migraphx::get_context($@)',
             const=True,
             returns='migraphx::context')


@auto_handle()
def operation(h):
    h.constructor('create',
                  api.params(name='const char*',
                             attributes='const char*',
                             vlist='...'),
                  fname='migraphx::create_op')
    h.method('name', returns='std::string')


api.add_function('migraphx_load',
                 api.params(name='const char*',
                            options='migraphx::file_options'),
                 fname='migraphx::load',
                 returns='migraphx::program')

api.add_function('migraphx_save',
                 api.params(p='migraphx::program&',
                            name='const char*',
                            options='migraphx::file_options'),
                 fname='migraphx::save')


@auto_handle()
def onnx_options(h):
    h.constructor('create')
    h.method(
        'set_input_parameter_shape',
        api.params(name='const char*', dims='std::vector<size_t>'),
        invoke='migraphx::set_input_parameter_shape($@)',
    )
    h.method(
        'set_dyn_input_parameter_shape',
        api.params(name='const char*',
                   dims='std::vector<migraphx::shape::dynamic_dimension>'),
        invoke='migraphx::set_dyn_input_parameter_shape($@)',
    )
    h.method(
        'set_default_dim_value',
        api.params(value='size_t'),
        invoke='migraphx::set_default_dim_value($@)',
    )
    h.method(
        'set_default_dyn_dim_value',
        api.params(dd='const migraphx::shape::dynamic_dimension&'),
        invoke='migraphx::set_default_dyn_dim_value($@)',
    )
    h.method(
        'set_default_loop_iterations',
        api.params(value='int64_t'),
        invoke='migraphx::set_default_loop_iterations($@)',
    )
    h.method(
        'set_limit_loop_iterations',
        api.params(value='int64_t'),
        invoke='migraphx::set_limit_loop_iterations($@)',
    )
    h.method(
        'set_external_data_path',
        api.params(external_data_path='const char*'),
        invoke='migraphx::set_external_data_path($@)',
    )


@auto_handle()
def file_options(h):
    h.constructor('create')
    h.method('set_file_format',
             api.params(format='const char*'),
             invoke='migraphx::set_file_format($@)')


@auto_handle()
def compile_options(h):
    h.constructor('create')
    h.method('set_offload_copy',
             api.params(value='bool'),
             invoke='migraphx::set_offload_copy($@)')
    h.method('set_fast_math',
             api.params(value='bool'),
             invoke='migraphx::set_fast_math($@)')
    h.method('set_exhaustive_tune_flag',
             api.params(value='bool'),
             invoke='migraphx::set_exhaustive_tune_flag($@)')


api.add_function('migraphx_parse_onnx',
                 api.params(name='const char*',
                            options='migraphx::onnx_options'),
                 fname='migraphx::parse_onnx',
                 returns='migraphx::program')

api.add_function('migraphx_parse_onnx_buffer',
                 api.params(data='const void*',
                            size='size_t',
                            options='migraphx::onnx_options'),
                 fname='migraphx::parse_onnx_buffer',
                 returns='migraphx::program')


@auto_handle()
def tf_options(h):
    h.constructor('create')
    h.method(
        'set_nhwc',
        api.params(is_nhwc='bool'),
        invoke='migraphx::set_nhwc($@)',
    )
    h.method(
        'set_input_parameter_shape',
        api.params(name='const char*', dims='std::vector<size_t>'),
        invoke='migraphx::set_input_parameter_shape($@)',
    )
    h.method(
        'set_default_dim_value',
        api.params(value='size_t'),
        invoke='migraphx::set_default_dim_value($@)',
    )
    h.method(
        'set_output_names',
        api.params(names='std::vector<const char*>'),
        invoke='migraphx::set_output_names($@)',
    )


api.add_function('migraphx_parse_tf',
                 api.params(name='const char*',
                            options='migraphx::tf_options'),
                 fname='migraphx::parse_tf',
                 returns='migraphx::program')

api.add_function('migraphx_parse_tf_buffer',
                 api.params(data='const void*',
                            size='size_t',
                            options='migraphx::tf_options'),
                 fname='migraphx::parse_tf_buffer',
                 returns='migraphx::program')


@api.handle('migraphx_quantize_op_names', 'std::vector<std::string>')
def quantize_op_names(h):
    h.constructor('create')
    h.method('add', api.params(name='const char*'), fname='push_back')


api.add_function('migraphx_quantize_fp16_with_op_names',
                 api.params(prog='migraphx::program&',
                            name='std::vector<std::string>&'),
                 fname='migraphx::quantize_fp16_with_op_names')

api.add_function('migraphx_quantize_fp16',
                 api.params(prog='migraphx::program&'),
                 fname='migraphx::quantize_fp16')

api.add_function('migraphx_quantize_bf16_with_op_names',
                 api.params(prog='migraphx::program&',
                            name='std::vector<std::string>&'),
                 fname='migraphx::quantize_bf16_with_op_names')

api.add_function('migraphx_quantize_bf16',
                 api.params(prog='migraphx::program&'),
                 fname='migraphx::quantize_bf16')


@auto_handle()
def quantize_int8_options(h):
    h.constructor('create')
    h.method(
        'add_op_name',
        api.params(name='const char*'),
        invoke='migraphx::add_op_name($@)',
    )
    h.method(
        'add_calibration_data',
        api.params(data='std::unordered_map<std::string, migraphx::argument>'),
        invoke='migraphx::add_calibration_data($@)',
    )


api.add_function('migraphx_quantize_int8',
                 api.params(prog='migraphx::program&',
                            target='migraphx::target',
                            options='migraphx::quantize_int8_options'),
                 fname='migraphx::quantize_int8_wrap')


@auto_handle()
def quantize_fp8_options(h):
    h.constructor('create')
    h.method(
        'add_calibration_data',
        api.params(data='std::unordered_map<std::string, migraphx::argument>'),
        invoke='migraphx::add_calibration_data($@)',
    )


api.add_function('migraphx_quantize_fp8',
                 api.params(prog='migraphx::program&',
                            target='migraphx::target',
                            options='migraphx::quantize_fp8_options'),
                 fname='migraphx::quantize_fp8_wrap')


@auto_handle(ref=True)
def context(h):
    h.method('finish', const=True)
    h.method('get_queue', returns='void*', fname='get_queue().unsafe_get')


@api.interface('migraphx_experimental_custom_op',
               'migraphx::experimental_custom_op')
def experimental_custom_op(h):
    h.constructor('create',
                  api.params(obj_typename='const char*', name='const char*'))
    h.virtual('compute',
              api.params(ctx='migraphx::context',
                         output='migraphx::shape',
                         inputs='std::vector<migraphx::argument>'),
              returns='migraphx::argument')
    h.virtual('compute_shape',
              api.params(inputs='std::vector<migraphx::shape>'),
              returns='migraphx::shape')
    h.virtual('output_alias',
              api.params(inputs='std::vector<migraphx::shape>'),
              returns='std::vector<size_t>')
    h.virtual('runs_on_offload_target', returns='bool')
    h.method('register', invoke='migraphx::register_custom_op($@)')



#####################################################################################
#  The MIT License (MIT)
#
#  Copyright (c) 2015-2025 Advanced Micro Devices, Inc. All rights reserved.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.
#####################################################################################
#################################### GUIDE ##########################################
#####################################################################################
# This license_stamper script is to be triggered manually via users to update the license
# stamps for all the files in the current local branch. It works by generating a list of
# all the files in the current active local branch via GIT LS-FILES and then compares each
# files license stamp year to the latest commit year (which is found via GIT LOG).
# It then updates the year if needed. If a license stamp is not found, then
# it will add a stamp at the begenning of the file with the year set to the current year.
#####################################################################################
import os
import argparse
from stamp_status import StampStatus, stamp_check, update_year, current_year
from git_tools import get_all_files, get_changed_files, get_latest_commit_year, get_top

delimiters = {
    ".cpp": "*",
    ".hpp": "*",
    ".h": "*",
    ".ipynb": "//",
    ".py": "#",
    ".txt": "#",
    ".bsh": "#",
    ".sh": "#",
    ".cmake": "#"
}
extensions = delimiters.keys()


# Get the delimiter from the file type based on what we care about to tag with our licence
# file. Otherwise return None for the delimiter and skip the file
def get_delimiter(filename):
    delimiter = None
    for extension in extensions:
        if filename.endswith(extension):
            delimiter = delimiters[extension]
            break
    return delimiter


def getipynb_markdown_block_aslist():
    markdown_block = [
        '\t{\n'
        '\t\t"cell_type": "code",\n', '\t\t"execution_count": null,\n',
        '\t\t"metadata": {},\n', '\t\t"outputs": [],\n', '\t\t"source": [\n',
        '\t\t\t\"#  The MIT License (MIT)\\n\",\n', '\t\t\t\"#\\n\",\n',
        f'\t\t\t\"#  Copyright (c) {current_year()} Advanced Micro Devices, Inc. All rights reserved.\\n\",\n',
        '\t\t\t\"#\\n\",\n',
        '\t\t\t\"#  Permission is hereby granted, free of charge, to any person obtaining a copy\\n\",\n',
        '\t\t\t\"#  of this software and associated documentation files (the \'Software\'), to deal\\n\",\n',
        '\t\t\t\"#  in the Software without restriction, including without limitation the rights\\n\",\n',
        '\t\t\t\"#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell\\n\",\n',
        '\t\t\t\"#  copies of the Software, and to permit persons to whom the Software is\\n\",\n',
        '\t\t\t\"#  furnished to do so, subject to the following conditions:\\n\",\n',
        '\t\t\t\"#\\n\",\n',
        '\t\t\t\"#  The above copyright notice and this permission notice shall be included in\\n\",\n',
        '\t\t\t\"#  all copies or substantial portions of the Software.\\n\",\n',
        '\t\t\t\"#\\n\",\n',
        '\t\t\t\"#  THE SOFTWARE IS PROVIDED \'AS IS\', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR\\n\",\n',
        '\t\t\t\"#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,\\n\",\n',
        '\t\t\t\"#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE\\n\",\n',
        '\t\t\t\"#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER\\n\",\n',
        '\t\t\t\"#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,\\n\",\n',
        '\t\t\t\"#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN\\n\",\n',
        '\t\t\t\"#  THE SOFTWARE.\\n\"\n', '\t\t]\n', '\t},'
    ]
    return markdown_block


# Header and footer of the comment block
# modify these if we want some different style
def top_header(comment_char):
    delim = None

    #Early return
    if "//" in comment_char:
        delim = getipynb_markdown_block_aslist()
        delim.append("\n")
        return ''.join(str(x) for x in delim)

    if "*" in comment_char:
        delim = "/*\n"
    if "#" in comment_char:
        delim = "#####################################################################################\n"
    return delim


def bottom_footer(comment_char):
    delim = None
    #Early return - no footer handled by
    if "//" in comment_char:
        return delim

    if "*" in comment_char:
        delim = "*/\n"
    if "#" in comment_char:
        delim = "#####################################################################################\n"
    return delim


# Simple just open and write stuff to each file with the license stamp
def stamp_file(rfile,
               message,
               comment_char,
               use_last_commit_year=False,
               debug=False):

    filename = os.path.join(get_top(), rfile)
    add_shebang = False
    #markdown file stamping for .ipynb
    save_markdown_lines = []
    modify_markdown = False

    #open save old contents and append things here
    if debug is True:
        print("Open", filename)

    try:
        with open(filename, 'r', encoding='utf-8') as contents:
            if comment_char != "//":
                saved_shebang = contents.readline()
                if "#!" in saved_shebang:
                    add_shebang = True
                else:
                    contents.seek(0)

            if comment_char == "//":
                save_markdown_lines.append(contents.readline())
                save_markdown_lines.append(contents.readline())
                modify_markdown = True

            save = contents.read()
    except (OSError, UnicodeDecodeError) as e:
        if debug:
            print(f"{e} \n Skipping file")
        return

    year = get_latest_commit_year(
        rfile) if use_last_commit_year else current_year()
    status = stamp_check(filename, year, save, debug=debug)

    if status in (StampStatus.OK, StampStatus.OTHER_LICENSE,
                  StampStatus.WRONG_YEAR):
        if status == StampStatus.WRONG_YEAR:
            update_year(filename, year)
        return

    if debug:
        print("Writing header")

    with open(filename, 'w', encoding='utf-8') as contents:
        if add_shebang:
            contents.write(saved_shebang + "\n")
        if modify_markdown:
            contents.write(''.join(save_markdown_lines))
        delim = top_header(comment_char)
        if delim:
            contents.write(delim)
        if not modify_markdown:
            for line in message:
                contents.write(f"{comment_char} {line}\n"
                               if line else f"{comment_char}\n")
        delim = bottom_footer(comment_char)
        if delim:
            contents.write(delim)
        contents.write(save)

    if debug is True:
        print("Done")


def main(args):
    with open(os.path.join(get_top(), 'LICENSE'), encoding='utf-8') as f:
        message = f.read().split('\n')

    filelist = None
    if args.all:
        filelist = get_all_files()
    else:
        filelist = get_changed_files(args.against)

    if args.debug is True:
        print("Target file list:\n" + '\n'.join(filelist))
        print("Output Message:\n" + '\n'.join(message))

    for rfile in filelist:
        comment_delim = get_delimiter(rfile)
        if comment_delim is not None:
            print(f"Updating file: {rfile}")
            stamp_file(rfile,
                       message,
                       comment_delim,
                       use_last_commit_year=args.all,
                       debug=args.debug)
        else:
            print(f"No valid delimeter for file: {rfile}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('against', default='origin/develop', nargs='?')
    parser.add_argument('-a',
                        '--all',
                        action='store_true',
                        help='Update all files')
    parser.add_argument('--debug',
                        action='store_true',
                        help='Enable debug output')
    args = parser.parse_args()
    main(args)



import toml
import torch
import argparse
from model_separable_rnnt import RNNT


def load_and_migrate_checkpoint(ckpt_path):
    checkpoint = torch.load(ckpt_path, map_location="cpu")
    migrated_state_dict = {}
    for key, value in checkpoint['state_dict'].items():
        key = key.replace("joint_net", "joint.net")
        migrated_state_dict[key] = value
    del migrated_state_dict["audio_preprocessor.featurizer.fb"]
    del migrated_state_dict["audio_preprocessor.featurizer.window"]
    return migrated_state_dict


def pytorch_rnnt_model(config_toml='configs/rnnt.toml',
                       checkpoint_path='rnnt.pt'):
    config = toml.load(config_toml)

    def add_blank_label(labels):
        if not isinstance(labels, list):
            raise ValueError("labels must be a list of symbols")
        labels.append("<BLANK>")
        return labels

    dataset_vocab = config['labels']['labels']
    rnnt_vocab = add_blank_label(dataset_vocab)
    featurizer_config = config['input_eval']
    model = RNNT(feature_config=featurizer_config,
                 rnnt=config['rnnt'],
                 num_classes=len(rnnt_vocab))

    model.load_state_dict(load_and_migrate_checkpoint(checkpoint_path))
    model.to('cuda')
    model.eval()

    return model


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--config_toml", default="configs/rnnt.toml")
    parser.add_argument("--checkpoint_path", default="rnnt.pt")
    args = parser.parse_args()
    pytorch_rnnt_model(args.mlcommons_inference_path, args.checkpoint_path)


if __name__ == "__main__":
    main()



# Modifications Copyright (C) 2022, Advanced Micro Devices, Inc. All rights reserved
# Copyright 2018 The Google AI Language Team Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Inference for squad/bert using onnx.

This is going to do the samem as 'python run_squad.py --do_predict=True ...' using a squad/bert model
that was converted to onnx. Lots of code was taken from run_squad.py.
You run it with:


python onnx_squad.py --model $SQUAD_MODEL/squad.onnx \
                     --vocab_file $BERT_BASE_DIR/uncased_L-12_H-768_A-12/vocab.txt
                     --predict_file $SQUAD_DATA/dev-v1.1.json \
                     --bert_config_file $BERT_BASE_DIR/uncased_L-12_H-768_A-12/bert_config.json \
                     --output /tmp/
"""

import argparse
import collections
import json
import math
import os
import sys
from timeit import default_timer as timer

import numpy as np
import onnxruntime as onnxrt
import six
from tokenizers import BertWordPieceTokenizer
from tokenizers import pre_tokenizers

RawResult = collections.namedtuple("RawResult",
                                   ["unique_id", "start_logits", "end_logits"])

Feature = collections.namedtuple("Feature", [
    "unique_id", "tokens", "example_index", "token_to_orig_map",
    "token_is_max_context"
])


class SquadExample(object):
    """A single training/test example for simple sequence classification."""
    def __init__(self,
                 qas_id,
                 question_text,
                 doc_tokens,
                 orig_answer_text=None,
                 start_position=None,
                 end_position=None):
        self.qas_id = qas_id
        self.question_text = question_text
        self.doc_tokens = doc_tokens
        self.orig_answer_text = orig_answer_text
        self.start_position = start_position
        self.end_position = end_position

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        s = []
        s.append("qas_id: %s" % (self.qas_id))
        s.append("question_text: %s" % (self.question_text))
        s.append("doc_tokens: [%s]" % (" ".join(self.doc_tokens)))
        if self.start_position:
            s.append("start_position: %d" % (self.start_position))
        if self.start_position:
            s.append("end_position: %d" % (self.end_position))
        return ", ".join(s)


def _check_is_max_context(doc_spans, cur_span_index, position):
    """Check if this is the 'max context' doc span for the token."""

    # Because of the sliding window approach taken to scoring documents, a single
    # token can appear in multiple documents. E.g.
    #  Doc: the man went to the store and bought a gallon of milk
    #  Span A: the man went to the
    #  Span B: to the store and bought
    #  Span C: and bought a gallon of
    #  ...
    #
    # Now the word 'bought' will have two scores from spans B and C. We only
    # want to consider the score with "maximum context", which we define as
    # the *minimum* of its left and right context (the *sum* of left and
    # right context will always be the same, of course).
    #
    # In the example the maximum context for 'bought' would be span C since
    # it has 1 left context and 3 right context, while span B has 4 left context
    # and 0 right context.
    best_score = None
    best_span_index = None
    for (span_index, doc_span) in enumerate(doc_spans):
        end = doc_span.start + doc_span.length - 1
        if position < doc_span.start:
            continue
        if position > end:
            continue
        num_left_context = position - doc_span.start
        num_right_context = end - position
        score = min(num_left_context,
                    num_right_context) + 0.01 * doc_span.length
        if best_score is None or score > best_score:
            best_score = score
            best_span_index = span_index

    return cur_span_index == best_span_index


def convert_examples_to_features(examples, tokenizer, max_seq_length,
                                 doc_stride, max_query_length):
    """Loads a data file into a list of `InputBatch`s."""

    res_input_ids = []
    res_input_mask = []
    res_segment_ids = []
    extra = []
    unique_id = 0

    for (example_index, example) in enumerate(examples):
        query_tokens = tokenizer.encode(example.question_text)

        if len(query_tokens) > max_query_length:
            query_tokens = query_tokens[0:max_query_length]

        tok_to_orig_index = []
        orig_to_tok_index = []
        all_doc_tokens = []
        for (i, token) in enumerate(example.doc_tokens):
            orig_to_tok_index.append(len(all_doc_tokens))
            sub_tokens = tokenizer.encode(token, add_special_tokens=False)
            for sub_token in sub_tokens.tokens:
                tok_to_orig_index.append(i)
                all_doc_tokens.append(sub_token)

        # The -3 accounts for [CLS], [SEP] and [SEP]
        max_tokens_for_doc = max_seq_length - len(query_tokens) - 3

        # We can have documents that are longer than the maximum sequence length.
        # To deal with this we do a sliding window approach, where we take chunks
        # of the up to our max length with a stride of `doc_stride`.
        _DocSpan = collections.namedtuple("DocSpan", ["start", "length"])
        doc_spans = []
        start_offset = 0
        while start_offset < len(all_doc_tokens):
            length = len(all_doc_tokens) - start_offset
            if length > max_tokens_for_doc:
                length = max_tokens_for_doc
            doc_spans.append(_DocSpan(start=start_offset, length=length))
            if start_offset + length == len(all_doc_tokens):
                break
            start_offset += min(length, doc_stride)

        for (doc_span_index, doc_span) in enumerate(doc_spans):
            tokens = []
            token_to_orig_map = {}
            token_is_max_context = {}
            segment_ids = []
            tokens.append("[CLS]")
            segment_ids.append(0)
            for token in query_tokens.tokens:
                tokens.append(token)
                segment_ids.append(0)
            tokens.append("[SEP]")
            segment_ids.append(0)

            for i in range(doc_span.length):
                split_token_index = doc_span.start + i
                token_to_orig_map[len(
                    tokens)] = tok_to_orig_index[split_token_index]

                is_max_context = _check_is_max_context(doc_spans,
                                                       doc_span_index,
                                                       split_token_index)
                token_is_max_context[len(tokens)] = is_max_context
                tokens.append(all_doc_tokens[split_token_index])
                segment_ids.append(1)
            tokens.append("[SEP]")
            segment_ids.append(1)

            input_ids = []
            for token in tokens:
                input_ids.append(tokenizer.token_to_id(token))

            # The mask has 1 for real tokens and 0 for padding tokens. Only real
            # tokens are attended to.
            input_mask = [1] * len(input_ids)

            # Zero-pad up to the sequence length.
            while len(input_ids) < max_seq_length:
                input_ids.append(0)
                input_mask.append(0)
                segment_ids.append(0)
            res_input_ids.append(np.array(input_ids, dtype=np.int64))
            res_input_mask.append(np.array(input_mask, dtype=np.int64))
            res_segment_ids.append(np.array(segment_ids, dtype=np.int64))
            feature = Feature(unique_id=unique_id,
                              tokens=tokens,
                              example_index=example_index,
                              token_to_orig_map=token_to_orig_map,
                              token_is_max_context=token_is_max_context)
            extra.append(feature)
            unique_id += 1
    return np.array(res_input_ids), np.array(res_input_mask), np.array(
        res_segment_ids), extra


def read_squad_examples(input_file):
    """Read a SQuAD json file into a list of SquadExample."""
    with open(input_file, "r") as f:
        input_data = json.load(f)["data"]

    def is_whitespace(c):
        if c == " " or c == "\t" or c == "\r" or c == "\n" or ord(c) == 0x202F:
            return True
        return False

    examples = []
    for idx, entry in enumerate(input_data):
        for paragraph in entry["paragraphs"]:
            paragraph_text = paragraph["context"]
            doc_tokens = []
            char_to_word_offset = []
            prev_is_whitespace = True
            for c in paragraph_text:
                if is_whitespace(c):
                    prev_is_whitespace = True
                else:
                    if prev_is_whitespace:
                        doc_tokens.append(c)
                    else:
                        doc_tokens[-1] += c
                    prev_is_whitespace = False
                char_to_word_offset.append(len(doc_tokens) - 1)

            for qa in paragraph["qas"]:
                qas_id = qa["id"]
                question_text = qa["question"]
                start_position = None
                end_position = None
                orig_answer_text = None
                example = SquadExample(qas_id=qas_id,
                                       question_text=question_text,
                                       doc_tokens=doc_tokens,
                                       orig_answer_text=orig_answer_text,
                                       start_position=start_position,
                                       end_position=end_position)
                examples.append(example)
    return examples


def write_predictions(all_examples, all_features, all_results, n_best_size,
                      max_answer_length, do_lower_case, output_prediction_file,
                      output_nbest_file):
    """Write final predictions to the json file."""
    example_index_to_features = collections.defaultdict(list)
    for feature in all_features:
        example_index_to_features[feature.example_index].append(feature)

    unique_id_to_result = {}
    for result in all_results:
        unique_id_to_result[result.unique_id] = result

    _PrelimPrediction = collections.namedtuple(  # pylint: disable=invalid-name
        "PrelimPrediction", [
            "feature_index", "start_index", "end_index", "start_logit",
            "end_logit"
        ])

    all_predictions = collections.OrderedDict()
    all_nbest_json = collections.OrderedDict()
    for (example_index, example) in enumerate(all_examples):
        features = example_index_to_features[example_index]
        prelim_predictions = []
        for (feature_index, feature) in enumerate(features):
            if not feature.unique_id in unique_id_to_result:
                print("feature not in unique_Id", feature.unique_id)
                continue
            result = unique_id_to_result[feature.unique_id]

            start_indexes = _get_best_indexes(result.start_logits, n_best_size)
            end_indexes = _get_best_indexes(result.end_logits, n_best_size)
            for start_index in start_indexes:
                for end_index in end_indexes:
                    # We could hypothetically create invalid predictions, e.g., predict
                    # that the start of the span is in the question. We throw out all
                    # invalid predictions.
                    if start_index >= len(feature.tokens):
                        continue
                    if end_index >= len(feature.tokens):
                        continue
                    if start_index not in feature.token_to_orig_map:
                        continue
                    if end_index not in feature.token_to_orig_map:
                        continue
                    if not feature.token_is_max_context.get(
                            start_index, False):
                        continue
                    if end_index < start_index:
                        continue
                    length = end_index - start_index + 1
                    if length > max_answer_length:
                        continue
                    prelim_predictions.append(
                        _PrelimPrediction(
                            feature_index=feature_index,
                            start_index=start_index,
                            end_index=end_index,
                            start_logit=result.start_logits[start_index],
                            end_logit=result.end_logits[end_index]))

        prelim_predictions = sorted(prelim_predictions,
                                    key=lambda x:
                                    (x.start_logit + x.end_logit),
                                    reverse=True)

        _NbestPrediction = collections.namedtuple(  # pylint: disable=invalid-name
            "NbestPrediction", ["text", "start_logit", "end_logit"])

        seen_predictions = {}
        nbest = []
        for pred in prelim_predictions:
            if len(nbest) >= n_best_size:
                break
            feature = features[pred.feature_index]

            tok_tokens = feature.tokens[pred.start_index:(pred.end_index + 1)]
            orig_doc_start = feature.token_to_orig_map[pred.start_index]
            orig_doc_end = feature.token_to_orig_map[pred.end_index]
            orig_tokens = example.doc_tokens[orig_doc_start:(orig_doc_end + 1)]
            tok_text = " ".join(tok_tokens)

            # De-tokenize WordPieces that have been split off.
            tok_text = tok_text.replace(" ##", "")
            tok_text = tok_text.replace("##", "")

            # Clean whitespace
            tok_text = tok_text.strip()
            tok_text = " ".join(tok_text.split())
            orig_text = " ".join(orig_tokens)

            final_text = get_final_text(tok_text, orig_text, do_lower_case)
            if final_text in seen_predictions:
                continue

            seen_predictions[final_text] = True
            nbest.append(
                _NbestPrediction(text=final_text,
                                 start_logit=pred.start_logit,
                                 end_logit=pred.end_logit))

        # In very rare edge cases we could have no valid predictions. So we
        # just create a nonce prediction in this case to avoid failure.
        if not nbest:
            nbest.append(
                _NbestPrediction(text="empty", start_logit=0.0, end_logit=0.0))

        assert len(nbest) >= 1

        total_scores = []
        for entry in nbest:
            total_scores.append(entry.start_logit + entry.end_logit)

        probs = _compute_softmax(total_scores)

        nbest_json = []
        for (i, entry) in enumerate(nbest):
            output = collections.OrderedDict()
            output["text"] = entry.text
            output["probability"] = probs[i]
            output["start_logit"] = float(entry.start_logit)
            output["end_logit"] = float(entry.end_logit)
            nbest_json.append(output)

        all_predictions[example.qas_id] = nbest_json[0]["text"]
        all_nbest_json[example.qas_id] = nbest_json

    with open(output_prediction_file, "w") as writer:
        writer.write(json.dumps(all_predictions, indent=4) + "\n")

    with open(output_nbest_file, "w") as writer:
        writer.write(json.dumps(all_nbest_json, indent=4) + "\n")


def get_final_text(pred_text, orig_text, do_lower_case):
    """Project the tokenized prediction back to the original text."""

    # When we created the data, we kept track of the alignment between original
    # (whitespace tokenized) tokens and our WordPiece tokenized tokens. So
    # now `orig_text` contains the span of our original text corresponding to the
    # span that we predicted.
    #
    # However, `orig_text` may contain extra characters that we don't want in
    # our prediction.
    #
    # For example, let's say:
    #   pred_text = steve smith
    #   orig_text = Steve Smith's
    #
    # We don't want to return `orig_text` because it contains the extra "'s".
    #
    # We don't want to return `pred_text` because it's already been normalized
    # (the SQuAD eval script also does punctuation stripping/lower casing but
    # our tokenizer does additional normalization like stripping accent
    # characters).
    #
    # What we really want to return is "Steve Smith".
    #
    # Therefore, we have to apply a semi-complicated alignment heruistic between
    # `pred_text` and `orig_text` to get a character-to-charcter alignment. This
    # can fail in certain cases in which case we just return `orig_text`.

    def _strip_spaces(text):
        ns_chars = []
        ns_to_s_map = collections.OrderedDict()
        for (i, c) in enumerate(text):
            if c == " ":
                continue
            ns_to_s_map[len(ns_chars)] = i
            ns_chars.append(c)
        ns_text = "".join(ns_chars)
        return (ns_text, ns_to_s_map)

    # We first tokenize `orig_text`, strip whitespace from the result
    # and `pred_text`, and check if they are the same length. If they are
    # NOT the same length, the heuristic has failed. If they are the same
    # length, we assume the characters are one-to-one aligned.
    tokenizer = pre_tokenizers.Sequence(
        [pre_tokenizers.Whitespace(),
         pre_tokenizers.Punctuation()])

    tok_text = []
    for item in tokenizer.pre_tokenize_str(orig_text):
        tok_text.append(item[0])

    tok_text = " ".join(tok_text)

    start_position = tok_text.find(pred_text)
    if start_position == -1:
        return orig_text
    end_position = start_position + len(pred_text) - 1

    (orig_ns_text, orig_ns_to_s_map) = _strip_spaces(orig_text)
    (tok_ns_text, tok_ns_to_s_map) = _strip_spaces(tok_text)

    if len(orig_ns_text) != len(tok_ns_text):
        return orig_text

    # We then project the characters in `pred_text` back to `orig_text` using
    # the character-to-character alignment.
    tok_s_to_ns_map = {}
    for (i, tok_index) in six.iteritems(tok_ns_to_s_map):
        tok_s_to_ns_map[tok_index] = i

    orig_start_position = None
    if start_position in tok_s_to_ns_map:
        ns_start_position = tok_s_to_ns_map[start_position]
        if ns_start_position in orig_ns_to_s_map:
            orig_start_position = orig_ns_to_s_map[ns_start_position]

    if orig_start_position is None:
        return orig_text

    orig_end_position = None
    if end_position in tok_s_to_ns_map:
        ns_end_position = tok_s_to_ns_map[end_position]
        if ns_end_position in orig_ns_to_s_map:
            orig_end_position = orig_ns_to_s_map[ns_end_position]

    if orig_end_position is None:
        return orig_text

    output_text = orig_text[orig_start_position:(orig_end_position + 1)]
    return output_text


def _get_best_indexes(logits, n_best_size):
    """Get the n-best logits from a list."""
    index_and_score = sorted(enumerate(logits),
                             key=lambda x: x[1],
                             reverse=True)
    best_indexes = []
    for i in range(len(index_and_score)):
        if i >= n_best_size:
            break
        best_indexes.append(index_and_score[i][0])
    return best_indexes


def _compute_softmax(scores):
    """Compute softmax probability over raw logits."""
    if not scores:
        return []

    max_score = None
    for score in scores:
        if max_score is None or score > max_score:
            max_score = score

    exp_scores = []
    total_sum = 0.0
    for score in scores:
        x = math.exp(score - max_score)
        exp_scores.append(x)
        total_sum += x

    probs = []
    for score in exp_scores:
        probs.append(score / total_sum)
    return probs


def main():
    parser = argparse.ArgumentParser(description='onnx squad')
    parser.add_argument('--model', required=True, help='model')
    parser.add_argument('--vocab_file', required=True, help='vocab_file')
    parser.add_argument('--bert_config_file', help='vocab_file')
    parser.add_argument('--predict_file', required=True, help='predict_file')
    parser.add_argument('--output_dir', help='output dir')
    parser.add_argument('--max_seq_length',
                        type=int,
                        default=256,
                        help='max_seq_length')
    parser.add_argument('--max_query_length',
                        type=int,
                        default=64,
                        help='max_query_length')
    parser.add_argument('--max_answer_length',
                        type=int,
                        default=30,
                        help='max_answer_length')
    parser.add_argument('--n_best_size',
                        type=int,
                        default=20,
                        help='n_best_size')
    parser.add_argument('--doc_stride',
                        type=int,
                        default=128,
                        help='doc_stride')
    parser.add_argument('--batch_size', type=int, default=1, help='batch_size')
    parser.add_argument('--profile',
                        action='store_true',
                        help='enable chrome timeline trace profiling.')
    parser.add_argument('--log', type=int, help='log level.')
    args = parser.parse_args()

    sess_options = None
    if args.profile:
        sess_options = onnxrt.SessionOptions()
        sess_options.enable_profiling = True
        sess_options.profile_file_prefix = os.path.basename(args.model)
    if args.log:
        sess_options = onnxrt.SessionOptions()
        sess_options.session_log_verbosity_level = args.log

    tokenizer = BertWordPieceTokenizer(args.vocab_file)

    eval_examples = read_squad_examples(input_file=args.predict_file)
    input_ids, input_mask, segment_ids, extra_data = \
        convert_examples_to_features(eval_examples, tokenizer, args.max_seq_length,
                                     args.doc_stride, args.max_query_length)

    sess = onnxrt.InferenceSession(args.model, sess_options)
    for input_meta in sess.get_inputs():
        print(input_meta)
    n = len(input_ids)
    bs = args.batch_size
    all_results = []
    start = timer()
    for idx in range(0, n, bs):
        data = {
            "input_ids:0": input_ids[idx:idx + bs],
            "input_mask:0": input_mask[idx:idx + bs],
            "segment_ids:0": segment_ids[idx:idx + bs]
        }
        result = sess.run(["unstack:0", "unstack:1"], data)
        in_batch = result[0].shape[1]
        for i in range(0, in_batch):
            unique_id = len(all_results)
            all_results.append(
                RawResult(unique_id=unique_id,
                          start_logits=result[0][0][i],
                          end_logits=result[1][0][i]))
            if unique_id > 0 and unique_id % 100 == 0:
                print("at {} {}sec per item".format(
                    unique_id, (timer() - start) / unique_id))
    end = timer()

    print("total time: {}sec, {}sec per item".format(
        end - start, (end - start) / len(all_results)))

    if args.output_dir:
        output_prediction_file = os.path.join(args.output_dir,
                                              "predictions.json")
        output_nbest_file = os.path.join(args.output_dir,
                                         "nbest_predictions.json")
        write_predictions(eval_examples, extra_data, all_results,
                          args.n_best_size, args.max_answer_length, True,
                          output_prediction_file, output_nbest_file)
    if args.profile:
        trace_file = sess.end_profiling()
        print("trace file written to: {}".format(trace_file))
    return 0


if __name__ == "__main__":
    sys.exit(main())



import pandas as pd
import re
import plotly.graph_objects as go
from plotly.subplots import make_subplots
import argparse


def read_compile_log(file_path):
    # Regex pattern to match lines with "value: __ ms"
    pattern = r'([\w: ]+):\s*([\d.]+(?:e[+-]?\d+)?)ms'

    data = []

    # Read the file and find matches
    with open(file_path, 'r') as file:
        for line in file:
            match = re.search(pattern, line)
            if match:
                operation = match.group(1).strip()
                try:
                    value = float(match.group(2))
                    data.append({'Operation': operation, 'Value': value})
                except:
                    print(f"Invalid Float: {match.group(2)}")

    return data


def filter_data(data, quantile):

    # Create a DataFrame from the list of dictionaries
    df = pd.DataFrame(data)
    df['Line Number'] = df.index
    df['Operation: Line Number'] = df.apply(
        lambda x: x["Operation"] + ": " + str(x['Line Number']), axis=1)
    df['Cumulative Sum'] = df['Value'].cumsum()

    print(
        f"Only included the top {round(1-quantile, 2)}% time-intensive operations"
    )
    top_operation_time = df['Value'].quantile(quantile)
    df = df[df['Value'] >= top_operation_time]

    return df


def plot_compile_analysis(df, output_path):
    df_sorted_by_time = df.sort_values(by='Value', ascending=False)
    df_sorted_by_total_time = df.groupby(
        'Operation')['Value'].sum().sort_values(ascending=False).reset_index()
    df_sorted_by_avg_time = df.groupby('Operation')['Value'].mean(
    ).sort_values(ascending=False).reset_index()
    df_counts = df['Operation'].value_counts().reset_index()

    fig = make_subplots(rows=2,
                        cols=4,
                        row_heights=[0.4, 0.6],
                        specs=[[{
                            'type': 'bar'
                        }, {
                            'type': 'bar'
                        }, {
                            'type': 'bar'
                        }, {
                            'type': 'bar'
                        }],
                               [{
                                   'type': 'scatter',
                                   'colspan': 4
                               }, None, None, None]],
                        subplot_titles=('Time Taken per Line',
                                        'Total Time Taken per Operation',
                                        'Avg Time Taken per Operation',
                                        'Number of times Operation called',
                                        'Compile Time Series Graph'),
                        vertical_spacing=0.25)

    fig.add_trace(go.Bar(
        x=df_sorted_by_time['Operation: Line Number'],
        y=df_sorted_by_time['Value'],
        name='Time Taken per Line',
        marker=dict(color='red'),
        hoverinfo='x+y',
    ),
                  row=1,
                  col=1)

    fig.add_trace(go.Bar(x=df_sorted_by_total_time['Operation'],
                         y=df_sorted_by_total_time['Value'],
                         name='Total Time Taken per Operation',
                         marker=dict(color='green'),
                         hoverinfo='x+y'),
                  row=1,
                  col=2)

    fig.add_trace(go.Bar(x=df_sorted_by_avg_time['Operation'],
                         y=df_sorted_by_avg_time['Value'],
                         name='Avg Time Taken per Operation',
                         marker=dict(color='green'),
                         hoverinfo='x+y'),
                  row=1,
                  col=3)

    fig.add_trace(go.Bar(x=df_counts['Operation'],
                         y=df_counts['count'],
                         name='Number of times Operation called',
                         marker=dict(color='royalblue'),
                         hoverinfo='x+y'),
                  row=1,
                  col=4)

    fig.update_xaxes(
        row=1,
        tickangle=45,
    )

    fig.add_trace(go.Scatter(
        x=list(range(len(df))),
        y=df['Cumulative Sum'],
        mode='lines+markers',
        text=df['Operation: Line Number'],
        hoverinfo='text',
        name='Cumulative Time',
        line=dict(width=2),
        marker=dict(size=3),
    ),
                  row=2,
                  col=1)

    fig.update_xaxes(
        title_text='Operation: Line Number',
        tickvals=list(range(len(df))),
        ticktext=df['Operation: Line Number'],
        row=2,
        col=1,
        tickangle=45,
    )

    fig.update_yaxes(title_text='Cumulative Time (ms)', row=2, col=1)

    fig.update_layout(
        title='Compile Time Analysis',
        title_x=0.5,
        title_font=dict(size=24, color='darkblue'),
        # margin=dict(l=0, r=0, t=100, b=50),
        uniformtext_minsize=12,
        uniformtext_mode='hide',
        height=1200,
        template='plotly_white',
        showlegend=False)

    fig.write_html(output_path)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('--file_path', required=True)
    parser.add_argument('--quantile', default=0.95)
    parser.add_argument('--output_path', default="compile_analysis.html")
    args = parser.parse_args()

    data = read_compile_log(args.file_path)
    df = filter_data(data, args.quantile)
    plot_compile_analysis(df, args.output_path)



#####################################################################################
# The MIT License (MIT)
#
# Copyright (c) 2015-2023 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#####################################################################################
import migraphx


def test_conv_relu():
    p = migraphx.parse_onnx("conv_relu_maxpool_test.onnx")
    print(p)
    print("Compiling ...")
    p.compile(migraphx.get_target("gpu"), offload_copy=False)
    print(p)
    params = {}

    for key, value in p.get_parameter_shapes().items():
        print("Parameter {} -> {}".format(key, value))
        params[key] = migraphx.to_gpu(migraphx.generate_argument(value))

    r = migraphx.from_gpu(p.run(params)[-1])
    print(r)


# TODO: placeholder until tuple shapes and arguments exposed
#def test_dyn_batch():
#    a = migraphx.shape.dynamic_dimension(1, 4, {2, 4})
#    b = migraphx.shape.dynamic_dimension(3, 3)
#    c = migraphx.shape.dynamic_dimension(32, 32)
#    dd_map = {"0": [a, b, c, c]}
#    p = migraphx.parse_onnx("conv_relu_maxpool_test.onnx",
#                            map_dyn_input_dims=dd_map)
#    print(p)
#    print("Compiling ...")
#    p.compile(migraphx.get_target("gpu"), offload_copy=False)
#
#    print(p)
#
#    def run_prog(batch_size):
#        params = {}
#        for key, value in p.get_parameter_shapes().items():
#            print("Parameter {} -> {}".format(key, value))
#            params[key] = migraphx.to_gpu(
#                migraphx.generate_argument(value.to_static(batch_size)))
#
#        print("before_output")
#        outputs = p.run(params)
#        print(outputs)
#        r = migraphx.from_gpu(p.run(params)[-1])
#        print(r)
#
#    run_prog(1)
#    run_prog(2)
#    run_prog(3)
#    run_prog(4)

test_conv_relu()



#####################################################################################
# The MIT License (MIT)
#
# Copyright (c) 2015-2024 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#####################################################################################

from txt2img import StableDiffusionMGX, get_args, image_to_tensor
import gradio as gr


def main():
    args = get_args()
    # Note: This will load the models, which can take several minutes
    sd = StableDiffusionMGX(args.onnx_model_path, args.compiled_model_path,
                            args.fp16, args.force_compile,
                            args.exhaustive_tune)
    sd.warmup(5)

    def gr_wrapper(prompt, negative_prompt, image, steps, seed, scale,
                   conditioning_scale):
        result = sd.run(str(prompt), str(negative_prompt),
                        image_to_tensor(image), int(steps), int(seed),
                        float(scale), float(conditioning_scale))
        return StableDiffusionMGX.convert_to_rgb_image(result)

    demo = gr.Interface(
        gr_wrapper,
        [
            gr.Textbox(value=args.prompt, label="Prompt"),
            gr.Textbox(value=args.negative_prompt,
                       label="Negative prompt (Optional)"),
            gr.Image(label="Canny Control Image",
                     value=args.control_image,
                     type='filepath'),
            gr.Slider(
                1, 100, step=1, value=args.steps, label="Number of steps"),
            gr.Textbox(value=args.seed, label="Random seed"),
            gr.Slider(
                1, 20, step=0.1, value=args.scale, label="Guidance scale"),
            gr.Slider(0,
                      1,
                      step=0.1,
                      value=args.conditioning_scale,
                      label="Conditioning scale"),
        ],
        "image",
    )
    demo.launch(share=True)


if __name__ == "__main__":
    main()



#####################################################################################
# The MIT License (MIT)
#
# Copyright (c) 2015-2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#####################################################################################
import numpy
import onnxruntime as rt

sess = rt.InferenceSession("dm_nfnet_f0.onnx")

input_name = sess.get_inputs()[0].name
print("input name", input_name)
input_shape = sess.get_inputs()[0].shape
print("input shape", input_shape)
input_type = sess.get_inputs()[0].type
print("input type", input_type)

output_name = sess.get_outputs()[0].name
print("output name", output_name)
output_shape = sess.get_outputs()[0].shape
print("output shape", output_shape)
output_type = sess.get_outputs()[0].type
print("output type", output_type)

x = numpy.random.random((1, 3, 192, 192))
x = x.astype(numpy.float32)

import migraphx

model = migraphx.parse_onnx("dm_nfnet_f0.onnx")
model.compile(migraphx.get_target("gpu"))
print(model.get_parameter_names())
print(model.get_parameter_shapes())
print(model.get_output_shapes())

result_migraphx = model.run({"inputs": x})
result_ort = sess.run([output_name], {input_name: x})

result_migraphx = result_migraphx[0].tolist()

for i in range(10):
    x = numpy.random.random((1, 3, 192, 192))
    x = x.astype(numpy.float32)

    result_migraphx = model.run({"inputs": x})
    result_ort = sess.run([output_name], {input_name: x})

    try:
        numpy.testing.assert_allclose(result_migraphx[0].tolist(),
                                      result_ort[0][0],
                                      rtol=1e-02)
        print(f"Test #{i} completed.")
    except AssertionError as e:
        print(e)



#####################################################################################
# The MIT License (MIT)
#
# Copyright (c) 2015-2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#####################################################################################

import os
import warnings
import time
from tabulate import tabulate
import numpy as np
import torch
from models import (get_tokenizer, get_clip_model, get_t5_model,
                    get_flux_transformer_model, get_vae_model, get_scheduler)
from PIL import Image
# import migraphx as mgx


def calculate_shift(
        image_seq_len,
        base_seq_len: int = 256,
        max_seq_len: int = 4096,
        base_shift: float = 0.5,
        max_shift: float = 1.16,
):
    m = (max_shift - base_shift) / (max_seq_len - base_seq_len)
    b = base_shift - m * base_seq_len
    mu = image_seq_len * m + b
    return mu


class FluxPipeline:
    def __init__(self,
                 hf_model_path="black-forest-labs/FLUX.1-dev",
                 local_dir=None,
                 compile_dir=None,
                 pipeline_type="txt2img",
                 img_height=1024,
                 img_width=1024,
                 guidance_scale=3.5,
                 max_sequence_length=512,
                 batch_size=1,
                 denoising_steps=50,
                 fp16=False,
                 bf16=True,
                 exhaustive_tune=False,
                 manual_seed=None):

        self.hf_model_path = hf_model_path
        self.height = img_height
        self.width = img_width
        self.guidance_scale = guidance_scale
        self.max_sequence_length = max_sequence_length
        self.pipeline_type = pipeline_type
        self.bs = batch_size
        self.steps = denoising_steps
        self.fp16 = fp16
        self.bf16 = bf16
        self.exhaustive_tune = exhaustive_tune

        if not local_dir:
            self.local_dir = self.hf_model_path.split("/")[-1]
        if not compile_dir:
            self.compile_dir = self.hf_model_path.split("/")[-1] + "_compiled"

        self.models = {}

        # self.stages = ["clip", "t5", "transformer", "vae"]

        self.generator = torch.Generator(device="cuda")
        if manual_seed:
            self.generator.manual_seed(manual_seed)
        self.device = torch.cuda.current_device()

        self.times = []

    def load_models(self):
        self.scheduler = get_scheduler(self.local_dir, self.hf_model_path)
        self.tokenizer = get_tokenizer(self.local_dir, self.hf_model_path)
        self.tokenizer2 = get_tokenizer(self.local_dir, self.hf_model_path,
                                        "t5", "tokenizer_2")

        self.clip = get_clip_model(self.local_dir,
                                   self.hf_model_path,
                                   self.compile_dir,
                                   fp16=self.fp16,
                                   bf16=self.bf16,
                                   bs=self.bs,
                                   exhaustive_tune=self.exhaustive_tune)

        self.t5 = get_t5_model(self.local_dir,
                               self.hf_model_path,
                               self.compile_dir,
                               self.max_sequence_length,
                               bs=self.bs,
                               bf16=self.bf16,
                               exhaustive_tune=self.exhaustive_tune)

        self.flux_transformer = get_flux_transformer_model(
            self.local_dir,
            self.hf_model_path,
            self.compile_dir,
            img_height=self.height,
            img_width=self.width,
            max_len=self.max_sequence_length,
            fp16=self.fp16,
            bf16=self.bf16,
            bs=self.bs,
            exhaustive_tune=self.exhaustive_tune)

        self.vae = get_vae_model(self.local_dir,
                                 self.hf_model_path,
                                 self.compile_dir,
                                 img_height=self.height,
                                 img_width=self.width,
                                 bs=self.bs,
                                 bf16=self.bf16,
                                 exhaustive_tune=self.exhaustive_tune)

    @staticmethod
    def _pack_latents(latents, batch_size, num_channels_latents, height,
                      width):
        """
        Reshapes latents from (B, C, H, W) to (B, H/2, W/2, C*4) as expected by the denoiser
        """
        latents = latents.view(batch_size, num_channels_latents, height // 2,
                               2, width // 2, 2)
        latents = latents.permute(0, 2, 4, 1, 3, 5)
        latents = latents.reshape(batch_size, (height // 2) * (width // 2),
                                  num_channels_latents * 4)

        return latents

    @staticmethod
    def _unpack_latents(latents, height, width, vae_scale_factor):
        """
        Reshapes denoised latents to the format (B, C, H, W)
        """
        batch_size, num_patches, channels = latents.shape

        height = height // vae_scale_factor
        width = width // vae_scale_factor

        latents = latents.view(batch_size, height, width, channels // 4, 2, 2)
        latents = latents.permute(0, 3, 1, 4, 2, 5)

        latents = latents.reshape(batch_size, channels // (2 * 2), height * 2,
                                  width * 2)

        return latents

    @staticmethod
    def _prepare_latent_image_ids(height, width, dtype, device):
        """
        Prepares latent image indices
        """
        latent_image_ids = torch.zeros(height // 2, width // 2, 3)
        latent_image_ids[..., 1] = (latent_image_ids[..., 1] +
                                    torch.arange(height // 2)[:, None])
        latent_image_ids[..., 2] = (latent_image_ids[..., 2] +
                                    torch.arange(width // 2)[None, :])

        latent_image_id_height, latent_image_id_width, latent_image_id_channels = (
            latent_image_ids.shape)

        latent_image_ids = latent_image_ids.reshape(
            latent_image_id_height * latent_image_id_width,
            latent_image_id_channels)

        return latent_image_ids.to(device=device, dtype=dtype)

    def initialize_latents(
            self,
            batch_size,
            num_channels_latents,
            latent_height,
            latent_width,
            latents_dtype=torch.float32,
    ):
        latents_dtype = latents_dtype  # text_embeddings.dtype
        latents_shape = (batch_size, num_channels_latents, latent_height,
                         latent_width)
        latents = torch.randn(
            latents_shape,
            device=torch.cuda.current_device(),
            dtype=latents_dtype,
            generator=self.generator,
        )

        latents = self._pack_latents(latents, batch_size, num_channels_latents,
                                     latent_height, latent_width)

        latent_image_ids = self._prepare_latent_image_ids(
            latent_height, latent_width, latents_dtype, self.device)

        return latents, latent_image_ids

    def encode_prompt(self,
                      prompt,
                      encoder="clip",
                      max_sequence_length=None,
                      pooled_output=False):
        tokenizer = self.tokenizer2 if encoder == "t5" else self.tokenizer
        encoder = self.t5 if encoder == "t5" else self.clip
        max_sequence_length = (tokenizer.model_max_length
                               if max_sequence_length is None else
                               max_sequence_length)

        def tokenize(prompt, max_sequence_length):
            text_input_ids = (tokenizer(
                prompt,
                padding="max_length",
                max_length=max_sequence_length,
                truncation=True,
                return_overflowing_tokens=False,
                return_length=False,
                return_tensors="pt",
            ).input_ids.type(torch.int32).to(self.device))

            untruncated_ids = tokenizer(prompt,
                                        padding="longest",
                                        return_tensors="pt").input_ids.type(
                                            torch.int32).to(self.device)
            if untruncated_ids.shape[-1] >= text_input_ids.shape[
                    -1] and not torch.equal(text_input_ids, untruncated_ids):
                removed_text = tokenizer.batch_decode(
                    untruncated_ids[:, max_sequence_length - 1:-1])
                warnings.warn(
                    "The following part of your input was truncated because `max_sequence_length` is set to "
                    f"{max_sequence_length} tokens: {removed_text}")

            # NOTE: output tensor for the encoder must be cloned because it will be overwritten when called again for prompt2
            outputs = encoder.run_async(input_ids=text_input_ids)
            output_name = ("main:#output_0"
                           if not pooled_output else "main:#output_1")
            text_encoder_output = outputs[output_name].clone()

            return text_encoder_output

        # Tokenize prompt
        text_encoder_output = tokenize(prompt, max_sequence_length)

        # return (text_encoder_output.to(torch.float16)
        #         if self.fp16 else text_encoder_output)
        return text_encoder_output

    def denoise_latent(
            self,
            latents,
            timesteps,
            text_embeddings,
            pooled_embeddings,
            text_ids,
            latent_image_ids,
            denoiser="transformer",
            guidance=None,
    ):

        # handle guidance
        if self.flux_transformer.config["guidance_embeds"] and guidance is None:
            guidance = torch.full([latents.shape[0]],
                                  self.guidance_scale,
                                  device=self.device,
                                  dtype=torch.float32)

        for step_index, timestep in enumerate(timesteps):
            # prepare inputs
            timestep_inp = timestep.expand(latents.shape[0]).to(latents.dtype)
            params = {
                "hidden_states": latents,
                "timestep": timestep_inp / 1000,
                "pooled_projections": pooled_embeddings,
                "encoder_hidden_states": text_embeddings,
                "txt_ids": text_ids,
                "img_ids": latent_image_ids,
            }
            if guidance is not None:
                params.update({"guidance": guidance})

            noise_pred = self.flux_transformer.run_async(
                **params)["main:#output_0"]

            latents = self.scheduler.step(noise_pred,
                                          timestep,
                                          latents,
                                          return_dict=False)[0]

        return latents.to(dtype=torch.float32)

    def decode_latent(self, latents):
        images = self.vae.run_async(latent=latents)["main:#output_0"]
        return images

    def infer(self, prompt, prompt2, warmup=False):
        assert len(prompt) == len(prompt2)
        batch_size = len(prompt)

        self.vae_scale_factor = 2**(len(self.vae.config["block_out_channels"]))
        latent_height = 2 * (int(self.height) // self.vae_scale_factor)
        latent_width = 2 * (int(self.width) // self.vae_scale_factor)

        num_inference_steps = self.steps

        with torch.inference_mode():
            torch.cuda.synchronize()

            self.e2e_tic = time.perf_counter()

            latents, latent_image_ids = self.initialize_latents(
                batch_size=batch_size,
                num_channels_latents=self.flux_transformer.
                config["in_channels"] // 4,
                # num_channels_latents=16,
                latent_height=latent_height,
                latent_width=latent_width,
                # latents_dtype=torch.float16 if self.fp16 else torch.float32,
                latents_dtype=torch.float32)

            pooled_embeddings = self.encode_prompt(prompt, pooled_output=True)
            text_embeddings = self.encode_prompt(
                prompt2,
                encoder="t5",
                max_sequence_length=self.max_sequence_length)
            text_ids = torch.zeros(text_embeddings.shape[1],
                                   3).to(device=self.device,
                                         dtype=text_embeddings.dtype)

            # Prepare timesteps
            sigmas = np.linspace(1.0, 1 / num_inference_steps,
                                 num_inference_steps)
            image_seq_len = latents.shape[1]
            mu = calculate_shift(
                image_seq_len,
                self.scheduler.config.base_image_seq_len,
                self.scheduler.config.max_image_seq_len,
                self.scheduler.config.base_shift,
                self.scheduler.config.max_shift,
            )
            self.scheduler.set_timesteps(sigmas=sigmas,
                                         mu=mu,
                                         device=self.device)
            timesteps = self.scheduler.timesteps.to(self.device)
            num_inference_steps = len(timesteps)

            latents = self.denoise_latent(
                latents,
                timesteps,
                text_embeddings,
                pooled_embeddings,
                text_ids,
                latent_image_ids,
            )

            latents = self._unpack_latents(latents, self.height, self.width,
                                           self.vae_scale_factor)
            latents = (latents / self.vae.config["scaling_factor"]
                       ) + self.vae.config["shift_factor"]

            images = self.decode_latent(latents)
            torch.cuda.synchronize()
            self.e2e_toc = time.perf_counter()
            if not warmup:
                self.record_times()

        return images

    def record_times(self):
        self.times.append(self.e2e_toc - self.e2e_tic)

    def save_image(self, images, prefix, output_dir="./"):
        images = ((images + 1) * 255 / 2).clamp(0, 255).detach().permute(
            0, 2, 3, 1).round().type(torch.uint8).cpu().numpy()
        for i in range(images.shape[0]):
            path = os.path.join(output_dir, f"{prefix}_{i}.png")
            Image.fromarray(images[i]).save(path)

    def print_summary(self):
        headers = ["Model", "Latency(ms)"]
        rows = []
        for mod in ("clip", "t5", "flux_transformer", "vae"):
            name = f"{mod} (x{self.steps})" if mod == "flux_transformer" else mod
            rows.append([name, np.average(getattr(self, mod).get_run_times())])

        rows.append(["e2e", np.average(self.times) * 1000])
        print(tabulate(rows, headers=headers))

    def clear_run_data(self):
        self.times = []
        for mod in (self.clip, self.t5, self.flux_transformer, self.vae):
            mod.clear_events()



#####################################################################################
# The MIT License (MIT)
#
# Copyright (c) 2015-2024 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#####################################################################################
import os
import onnxruntime as ort
import numpy as np
import torch

from .generic import _inference
from ..utils import get_model_io


class DiffusionStage(object):
    def __init__(self, model_path):
        self.model_path = model_path
        self._initilize()

    def _initilize(self):
        self.inputs, self.outputs = get_model_io(self.model_path)
        self.session = ort.InferenceSession(self.model_path)
        self.folder_name_prefix = f"{os.path.dirname(self.model_path)}/test_data_set"

    def inference(self, input_data_map, test_idx):
        folder_name = f"{self.folder_name_prefix}_{test_idx}"
        output_data_map = _inference(
            self.session,
            input_data_map,
            self.outputs,
            folder_name,
        )
        return output_data_map


def generate_diffusion_data(model,
                            image_dataset,
                            prompt_dataset,
                            output_folder_prefix=None,
                            sample_limit=None,
                            decode_limit=None):
    assert model.is_diffuser() == True, "Only diffuser supported"
    output_path = f"{output_folder_prefix or 'generated'}/diffusion/{model.name()}"

    try:
        model_paths = model.get_model(output_path)
    except Exception as e:
        print(f"Something went wrong:\n{e}\nSkipping model...")
        return

    is_xl = 'xl' in model.name()
    assert len(model_paths) == (4 + is_xl)
    text_encoder = DiffusionStage(model_paths[0])
    text_encoder_2 = DiffusionStage(model_paths[1]) if is_xl else None
    vae_encoder = DiffusionStage(model_paths[1 + is_xl])
    unet = DiffusionStage(model_paths[2 + is_xl])
    vae_decoder = DiffusionStage(model_paths[3 + is_xl])

    print('\n'.join([
        f"Creating {stage.folder_name_prefix}s..." for stage in
        [text_encoder, text_encoder_2, vae_encoder, unet, vae_decoder]
        if stage is not None
    ]))

    test_idx = 0
    scale = 7.0
    seed = 42
    # TODO: get it from latent size
    size = 128 if is_xl else 64
    sample_shape = (1, 4, size, size)
    # TODO from config?
    vae_scaling_factor = 0.18215
    noise = torch.randn(sample_shape, generator=torch.manual_seed(seed))
    time_ids = np.array([[size * 8, size * 8, 0, 0, size * 8, size * 8]] * 2,
                        dtype=np.float32)

    for idx, (image, prompt) in enumerate(zip(image_dataset, prompt_dataset)):
        text_encoder_input_data_map = prompt_dataset.transform(
            text_encoder.inputs, prompt, model.text_preprocess)
        text_encoder_output_data_map = text_encoder.inference(
            text_encoder_input_data_map, test_idx)

        if is_xl:
            text_encoder_2_input_data_map = prompt_dataset.transform(
                text_encoder_2.inputs, prompt, model.text_preprocess_2)
            text_encoder_2_output_data_map = text_encoder_2.inference(
                text_encoder_2_input_data_map, test_idx)
            text_encoder_output_data_map['last_hidden_state'] = np.concatenate(
                (text_encoder_output_data_map['last_hidden_state'],
                 text_encoder_2_output_data_map['last_hidden_state']),
                axis=2)

        vae_encoder_input_data_map = image_dataset.transform(
            vae_encoder.inputs, image, model.image_preprocess)
        vae_encoder_output_data_map = vae_encoder.inference(
            vae_encoder_input_data_map, test_idx)

        model.scheduler.set_timesteps(decode_limit)
        latents = torch.from_numpy(
            vae_encoder_output_data_map["latent_sample"]) * vae_scaling_factor
        latents = model.scheduler.add_noise(latents, noise,
                                            model.scheduler.timesteps[:1])

        for step, t in enumerate(model.scheduler.timesteps):
            latents_model_input = torch.concatenate([latents] * 2)
            latents_model_input = model.scheduler.scale_model_input(
                latents_model_input, t).numpy().astype(np.float32)
            timestep = np.atleast_1d(t.numpy().astype(np.int64))

            unet_input_data_map = {
                'sample':
                latents_model_input,
                'encoder_hidden_states':
                text_encoder_output_data_map['last_hidden_state'],
                'timestep':
                timestep
            }
            if is_xl:
                unet_input_data_map[
                    'text_embeds'] = text_encoder_2_output_data_map[
                        'text_embeds']
                unet_input_data_map['time_ids'] = time_ids

            unet_output_data_map = unet.inference(
                unet_input_data_map, test_idx * decode_limit + step)

            noise_pred_text, noise_pred_uncond = np.array_split(
                unet_output_data_map['out_sample'], 2)

            noise_pred = noise_pred_uncond + scale * (noise_pred_text -
                                                      noise_pred_uncond)

            # compute the previous noisy sample x_t -> x_t-1
            latents = model.scheduler.step(torch.from_numpy(noise_pred), t,
                                           latents).prev_sample

        latents = 1 / vae_scaling_factor * latents
        vae_decoder_input_data_map = {
            "latent_sample": latents.numpy().astype(np.float32)
        }
        _ = vae_decoder.inference(vae_decoder_input_data_map, test_idx)

        test_idx += 1
        if sample_limit and sample_limit - 1 <= idx:
            break



#!/usr/bin/env python3

#####################################################################################
# The MIT License (MIT)
#
# Copyright (c) 2015-2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#####################################################################################

import json
import argparse
import os
from sys import argv as sysargs
from sys import version_info as python_version
from sys import exit as sys_exit
import pandas as pd
from datetime import datetime
import venv
import shutil

if (python_version[0] < 3) or (python_version[0] < 3
                               and python_version[1] < 6):
    raise Exception("Please utilize Python version 3.6 and above. Exiting...")


def parse_args():
    parser = argparse.ArgumentParser(
        description="Parser for MIGraphX ROCTX Markers")
    parser.add_argument('--json-path',
                        type=str,
                        metavar='json-path',
                        help='Path to json file')
    parser.add_argument('--out',
                        type=str,
                        metavar='out',
                        help='Output directory for run.')
    parser.add_argument(
        '--study-name',
        type=str,
        metavar='study-name',
        help='Study-name is used for naming the output CSV file.')
    parser.add_argument('--repeat',
                        type=int,
                        metavar='repeat',
                        help='Defines number of runs.',
                        default=2)
    parser.add_argument('--parse',
                        default=False,
                        action='store_true',
                        help='Parses given JSON file.')
    parser.add_argument('--clean',
                        default=False,
                        action='store_true',
                        help='Removes temporary paths')
    parser.add_argument('--run',
                        type=str,
                        metavar='run',
                        help='Enables run and fetches run configs.')
    parser.add_argument('--debug', default=False, action='store_true')

    args = parser.parse_args()
    return args


args = parse_args()
if not len(sysargs) > 1:
    raise Exception("No arg is passed. Exiting...")


def parse(file):
    with open(file, "r") as read_file:
        data = json.load(read_file)

    #Get marker names and first marker's time
    list_names = []
    first_marker = True
    first_marker_time = 0
    for i in data:
        if (i):
            if ("Marker start:" in i['name']) and (
                    i['name'] not in list_names):
                list_names.append(i['name'])
                if first_marker:
                    first_marker_time = i['ts']
                    first_marker = False

    if (args.debug):
        print(f"FIRST MARKER TIME DETERMINED: {first_marker_time}")

    if (first_marker_time == 0):
        raise ("FIRST MARKER TIME IS ZERO. EXITING...")

    kernel_launch_info = []  #kernel description
    kernel_launch_list = []  #kernel launch details
    kernel_launch_time = []  #kernel execution time
    for i in data:
        if (i and i.get('args')):
            try:
                if (("KernelExecution" in i['args']['desc'])
                        and (i['ts'] >= first_marker_time)):
                    kernel_launch_info.append(i['args']['desc'])
                    kernel_launch_list.append(i)
                    kernel_launch_time.append(int(i['dur']))
            except:
                continue

    max_index = kernel_launch_time.index(max(kernel_launch_time))
    max_kernel_info = kernel_launch_list[max_index]

    if (args.debug):
        with open('rocTX_kernel_launch_list.txt', 'w') as f:
            for i in kernel_launch_list:
                f.write(f'{i}')

    # Get timing information for each marker name
    list_times_per_names = []
    for name in list_names:
        temp_list = []
        for entry in data:
            if (entry) and (
                    name == entry['name']
            ):  # name can match on gpu or cpu side, for gpu, we need data from gpu markers.
                if (("gpu::" in name)
                        and ("UserMarker frame:" in entry['args']['desc'])
                    ):  #gpu side information
                    temp_list.append(int(entry.get('dur')))
                elif (("gpu::" not in name)
                      and ("Marker start:" in entry['args']['desc'])
                      ):  #cpu side information
                    temp_list.append(int(entry.get('dur')))
        list_times_per_names.append(temp_list)

    if (args.debug):
        print(list_times_per_names)

    sum_per_name = []  #TODO: refactor stat collection
    for list in list_times_per_names:
        sum_per_name.append(sum(list))

    count_per_name = []
    for list in list_times_per_names:
        try:
            count_per_name.append(len(list))
        except:
            count_per_name.append(0)

    max_per_name = []
    for list in list_times_per_names:
        try:
            max_per_name.append(max(list))
        except:
            max_per_name.append(0)

    min_per_name = []
    for list in list_times_per_names:
        try:
            min_per_name.append(min(list))
        except:
            min_per_name.append(0)

    max_index_per_name = []
    for list in list_times_per_names:
        try:
            max_index_per_name.append(list.index(max(list)))
        except:
            max_index_per_name.append(0)

    max_occur_per_name = []
    for list in list_times_per_names:
        try:
            max_occur_per_name.append(list.count(max(list)))
        except:
            max_occur_per_name.append(0)

    total_time = sum(sum_per_name)

    d = {
        'SUM': sum_per_name,
        'MIN': min_per_name,
        'MAX': max_per_name,
        'COUNT': count_per_name,
        'MAX_INDEX': max_index_per_name,
        'MAX_OCCUR': max_occur_per_name
    }
    df2 = pd.DataFrame(d)
    df2.index = list_names
    df2.sort_values(by=['SUM'], inplace=True, ascending=False)

    if (args.debug):
        print(df2)
        print(f"\nTOTAL TIME: {total_time} us")
    return df2, total_time, max_kernel_info


def run():
    repeat_count = args.repeat
    if (repeat_count == 0 or repeat_count == float('inf') or not repeat_count):
        raise Exception("REPEAT COUNT CANNOT BE ZERO/INFINITY/NULL")
    run_args = args.run
    #configurations
    configs = '--hip-trace --roctx-trace --flush-rate 10ms --timestamp on'
    output_dir = f"-d {args.out}"
    executable = f"/opt/rocm/bin/migraphx-driver roctx {run_args}"
    process_args = configs + ' ' + output_dir + ' ' + executable
    for i in range(repeat_count):
        os.system('rocprof ' + process_args)
    print("RUN COMPLETE.")


def clean():
    shutil.rmtree('/tmp/rocm-profile-data/', ignore_errors=False)


def main():

    if (args.clean):
        clean()
        sys_exit()

    print("Initiating virtual environment...")
    builder = venv.EnvBuilder(clear=True, with_pip=True)
    builder.create('/tmp/rocm-profile-data/py/')
    python_bin = '/tmp/rocm-profile-data/py' + '/bin/python'
    file = args.json_path

    if (args.study_name):
        filename = args.study_name + ".csv"
    else:
        filename = "output" + datetime.now().strftime(
            "%Y_%m_%d-%I:%M:%S_%p") + ".csv"

    with open(filename, 'a') as f:
        f.write(f"{args.run}\n")

    if (args.run):
        curr = os.path.abspath(os.getcwd())
        rpd_path = '/tmp/rocm-profile-data/rocmProfileData/'
        if not os.path.exists(rpd_path):
            print("rocmProfileData DOES NOT EXIST. CLONING...")
            os.system(
                f"git clone https://github.com/ROCmSoftwarePlatform/rocmProfileData.git {rpd_path}"
            )
        os.chdir(rpd_path + "rocpd_python/")
        os.system(python_bin + ' -m pip install --upgrade pip')
        os.system(python_bin + ' setup.py install')
        os.chdir(curr)
        run()
        os.chdir(curr + f"/{args.out}/")
        out_path = os.popen(f"ls -td $PWD/*/*/ | head -{args.repeat}").read()
        print(f"\nFOLLOWING PATHS WILL BE PARSED:\n{out_path}")
        out_path = out_path.splitlines()
        df_tot = pd.DataFrame()
        tot_time = []
        max_kernel_info_list = []
        for path in out_path:
            path = path.strip('\n')
            print("\nPARSING OUTPUT PATH: " + path)
            os.chdir(path)
            os.system(
                f"{python_bin} -m rocpd.rocprofiler_import --ops_input_file hcc_ops_trace.txt --api_input_file hip_api_trace.txt --roctx_input_file roctx_trace.txt trace.rpd"
            )
            os.system(
                f"{python_bin} {rpd_path}/rpd2tracing.py trace.rpd trace.json")
            os.chdir(curr)
            df, total_time, path_max_kernel_info = parse(path + "trace.json")
            max_kernel_info_list.append(path_max_kernel_info)
            tot_time.append(total_time)
            df_tot = pd.merge(df_tot,
                              df,
                              how='outer',
                              left_index=True,
                              right_index=True)
            if (args.debug):
                print("JSON FILE PATH: " + path + "trace.json")

        df_tot.to_csv("rocTX_runs_dataframe.csv")
        if (args.debug):
            print(df_tot)

        tmp_sum = df_tot.loc[:, df_tot.columns.str.contains('SUM')].astype(int)
        tmp_min = df_tot.loc[:, df_tot.columns.str.contains('MIN')].astype(int)
        tmp_max = df_tot.loc[:, df_tot.columns.str.match("^MAX_.$")].astype(
            int)
        tmp_count = df_tot.loc[:, df_tot.columns.str.match("COUNT")].astype(
            int)

        tmp_sum['SUM_avg'] = tmp_sum.mean(axis=1).astype(int)
        tmp_min['MIN_avg'] = tmp_min.mean(axis=1).astype(int)
        tmp_max['MAX_avg'] = tmp_max.mean(axis=1).astype(int)

        df2 = tmp_sum['SUM_avg'].copy()
        df2 = pd.merge(df2,
                       tmp_min['MIN_avg'],
                       how='outer',
                       left_index=True,
                       right_index=True)
        df2 = pd.merge(df2,
                       tmp_max['MAX_avg'],
                       how='outer',
                       left_index=True,
                       right_index=True)
        df2 = pd.merge(df2,
                       tmp_count['COUNT_x'],
                       how='outer',
                       left_index=True,
                       right_index=True)
        df2.rename(columns={'COUNT_x': 'COUNT'}, inplace=True)
        df2 = df2.loc[:, ~df2.columns.duplicated(
        )]  #there will be many COUNT_x in df2
        df2.sort_values(by=['SUM_avg'], inplace=True, ascending=False)

        if (args.debug):
            pd.set_option('display.max_columns', None)
            print(df_tot)  #all data from all runs

        print("\n*** RESULTS ***")
        print(df2)
        out_time = sum(tot_time) / len(tot_time)
        print(f"\nAVG TOTAL TIME: {out_time} us\n")

        df2.to_csv(filename, mode='a')
        with open(filename, 'a') as f:
            f.write(f"AVG TOTAL TIME: {out_time} us\n")
        print(f"OUTPUT CSV FILE:\t{filename}")

        if (args.debug):
            #kernels that took the longest time printed
            for item in max_kernel_info_list:
                print(f"KERNEL NAME: {item['name']}\t\t{item['dur']}")

        with open('rocTX_kernel_timing_details.txt', 'w') as f:
            f.write(
                "MOST TIME CONSUMING KERNELS IN EACH ITERATION (EXPECTED TO BE SAME KERNEL):\n"
            )
            for i in max_kernel_info_list:
                f.write(f"KERNEL NAME: {i['name']}\t\t{i['dur']}\n")
        print("KERNEL TIMING DETAILS:\trocTX_kernel_timing_details.txt")
        print("ALL DATA FROM ALL RUNS:\trocTX_runs_dataframe.csv")

    elif (args.parse):
        if not (file):
            raise Exception("JSON PATH IS NOT PROVIDED FOR PARSING.")
        parse(file)
    else:
        raise Exception("PLEASE PROVIDE A COMMAND: RUN, PARSE, CLEAN")


if __name__ == "__main__":
    main()
